from collections.abc import Mapping
from typing import TypeAlias

from sari.mcp.tools._util import (
    mcp_response,
    pack_header,
    pack_line,
    pack_encode_id,
    pack_encode_text,
    resolve_root_ids,
    resolve_repo_scope,
    pack_error,
    ErrorCode,
    parse_int_arg,
    invalid_args_response,
)
from sari.core.services.symbol_service import SymbolService

ToolResult: TypeAlias = dict[str, object]


def execute_get_implementations(args: object, db: object, roots: list[str]) -> ToolResult:
    """
    SymbolService를 사용하여 특정 심볼을 구현(implements)하거나 상속(extends)하는 심볼들을 찾습니다.
    (Interface Implementation / Subclass Search)
    """
    if not isinstance(args, Mapping):
        return invalid_args_response("get_implementations", "args must be an object")

    target_symbol = str(args.get("name", "")).strip()
    target_sid = str(args.get("symbol_id", "")).strip() or str(args.get("sid", "")).strip()
    target_path = str(args.get("path", "")).strip()
    repo = str(args.get("repo", "")).strip()
    limit, err = parse_int_arg(args, "limit", 100, "get_implementations", min_value=1, max_value=500)
    if err:
        return err
    if limit is None:
        return invalid_args_response("get_implementations", "'limit' must be an integer")
    
    if not target_symbol and not target_sid:
        return mcp_response(
            "get_implementations",
            lambda: pack_error("get_implementations", ErrorCode.INVALID_ARGS, "Symbol name or symbol_id is required"),
            lambda: {"error": {"code": ErrorCode.INVALID_ARGS.value, "message": "Symbol name or symbol_id is required"}, "isError": True},
        )

    # 1. 유효 범위(Scope) 해결
    allowed_root_ids = resolve_root_ids(roots)
    req_root_ids = args.get("root_ids")
    if isinstance(req_root_ids, list) and req_root_ids:
        req_set = {str(x) for x in req_root_ids if x}
        effective_root_ids = [rid for rid in allowed_root_ids if rid in req_set]
    else:
        effective_root_ids = allowed_root_ids

    _, repo_root_ids = resolve_repo_scope(repo, roots, db=db)
    if repo_root_ids:
        repo_set = set(repo_root_ids)
        effective_root_ids = [rid for rid in effective_root_ids if rid in repo_set] if effective_root_ids else list(repo_root_ids)

    # 2. 서비스 계층 위임 (순수 비즈니스 로직)
    service = SymbolService(db)
    results = service.get_implementations(
        target_name=target_symbol,
        symbol_id=target_sid,
        path=target_path,
        limit=limit,
        root_ids=effective_root_ids
    )

    def build_pack() -> str:
        header_params = {
            "name": pack_encode_text(target_symbol),
            "sid": pack_encode_id(target_sid),
            "path": pack_encode_id(target_path),
            "repo": pack_encode_id(repo)
        }
        lines = [pack_header("get_implementations", header_params, returned=len(results))]
        for r in results:
            kv = {
                "implementer_path": pack_encode_id(r["implementer_path"]),
                "implementer_symbol": pack_encode_id(r["implementer_symbol"]),
                "implementer_sid": pack_encode_id(r.get("implementer_sid", "")),
                "rel_type": pack_encode_id(r["rel_type"]),
                "line": str(r["line"]),
            }
            lines.append(pack_line("r", kv))
        return "\n".join(lines)

    return mcp_response(
        "get_implementations",
        build_pack,
        lambda: {"target": target_symbol, "target_sid": target_sid, "results": results, "count": len(results)},
    )



import time
import threading
from typing import Callable, Optional

class TokenBucket:
    """Rate limiting using Token Bucket algorithm."""
    def __init__(self, capacity: float, fill_rate: float):
        self.capacity = capacity
        self.tokens = capacity
        self.fill_rate = fill_rate
        self.last_update = time.time()
        self._lock = threading.Lock()

    def consume(self, amount: float = 1.0) -> bool:
        with self._lock:
            now = time.time()
            # Refill
            self.tokens = min(self.capacity, self.tokens + (now - self.last_update) * self.fill_rate)
            self.last_update = now
            
            if self.tokens >= amount:
                self.tokens -= amount
                return True
            return False

class AdaptiveDebouncer:
    """
    Debounces events with an adaptive window.
    Window increases when event burst is detected.
    """
    def __init__(self, callback: Callable, min_delay: float = 0.5, max_delay: float = 5.0):
        self.callback = callback
        self.min_delay = min_delay
        self.max_delay = max_delay
        self.current_delay = min_delay
        self._timer: Optional[threading.Timer] = None
        self._last_event_time = 0.0
        self._event_count_in_window = 0
        self._lock = threading.Lock()

    def handle_event(self, *args, **kwargs):
        with self._lock:
            now = time.time()
            elapsed = now - self._last_event_time
            self._last_event_time = now
            
            # If events come too fast, increase delay
            if elapsed < self.min_delay:
                self.current_delay = min(self.max_delay, self.current_delay * 1.5)
            else:
                self.current_delay = max(self.min_delay, self.current_delay * 0.8)

            if self._timer:
                self._timer.cancel()
            
            self._timer = threading.Timer(self.current_delay, self.callback, args=args, kwargs=kwargs)
            self._timer.start()



from sari.core.daemon_health import detect_orphan_daemons


class _Proc:
    def __init__(self, pid: int, cmdline: list[str]):
        self.info = {"pid": pid, "cmdline": cmdline}


def test_detect_orphan_daemons_filters_registered_pids(monkeypatch):
    monkeypatch.setattr(
        "sari.core.daemon_health._get_registry_daemon_pids",
        lambda: {100},
    )
    monkeypatch.setattr(
        "sari.core.daemon_health.psutil",
        type(
            "P",
            (),
            {
                "process_iter": staticmethod(
                    lambda *_args, **_kwargs: [
                        _Proc(100, ["python", "-m", "sari.mcp.daemon"]),
                        _Proc(200, ["python", "-m", "sari.mcp.daemon"]),
                        _Proc(300, ["python", "-m", "something_else"]),
                    ]
                )
            },
        )(),
    )

    result = detect_orphan_daemons()
    assert len(result) == 1
    assert result[0]["pid"] == 200



import json
import os
import signal
import threading
import time
import ipaddress
from pathlib import Path
from datetime import datetime
import sys
from sari.core.settings import settings

# Support both `python3 app/main.py` (script mode) and package mode.
try:
    from .config import Config, resolve_config_path  # type: ignore
    from .db import LocalSearchDB  # type: ignore
    from .http_server import serve_forever  # type: ignore
    from .indexer import Indexer  # type: ignore
    from .workspace import WorkspaceManager  # type: ignore
except ImportError:  # script mode
    from config import Config, resolve_config_path  # type: ignore
    from db import LocalSearchDB  # type: ignore
    from http_server import serve_forever  # type: ignore
    from indexer import Indexer  # type: ignore
    from workspace import WorkspaceManager  # type: ignore


def _resolve_http_host(cfg_host: str, allow_non_loopback: bool) -> str:
    host = (cfg_host or "127.0.0.1").strip()
    try:
        is_loopback = host.lower() == "localhost" or ipaddress.ip_address(host).is_loopback
    except ValueError:
        # Non-IP hostnames are only allowed if they resolve to localhost explicitly.
        is_loopback = host.lower() == "localhost"
    if (not is_loopback) and (not allow_non_loopback):
        raise SystemExit(
            f"sari refused to start: server_host must be loopback only (127.0.0.1/localhost/::1). got={host}. "
            "Set SARI_ALLOW_NON_LOOPBACK=1 to override (NOT recommended)."
        )
    return host


def _resolve_version() -> str:
    try:
        from sari.version import __version__
        return __version__
    except Exception:
        return os.environ.get("SARI_VERSION", "dev")


def _create_mcp_server(workspace_root: str, cfg: Config, db: LocalSearchDB, indexer: Indexer):
    try:
        from sari.mcp.server import LocalSearchMCPServer
        return LocalSearchMCPServer(workspace_root, cfg=cfg, db=db, indexer=indexer)
    except Exception:
        return None


def _write_server_info(workspace_root: str, host: str, actual_port: int, config_port: int) -> None:
    data_dir = Path(workspace_root) / ".codex" / "tools" / "sari" / "data"
    data_dir.mkdir(parents=True, exist_ok=True)
    server_json = data_dir / "server.json"
    server_info = {
        "host": host,
        "port": actual_port,  # use actual bound port, not config port
        "config_port": config_port,  # original requested port for reference
        "pid": os.getpid(),
        "started_at": datetime.now().isoformat(),
    }
    server_json.write_text(json.dumps(server_info, indent=2), encoding="utf-8")


def main() -> int:
    # Auto-detect workspace root for HTTP fallback
    WorkspaceManager.set_settings(settings)
    workspace_root = WorkspaceManager.resolve_workspace_root()

    # Set env var so Config can pick it up
    os.environ["SARI_WORKSPACE_ROOT"] = workspace_root

    cfg_path = resolve_config_path(workspace_root)

    # Graceful config loading (Global Install Support)
    try:
        cfg = Config.load(cfg_path, workspace_root_override=workspace_root)
    except Exception as e:
        # Fail fast with an actionable error to avoid ambiguous startup hangs.
        print(f"[sari] Config load failed ({cfg_path}): {e}", file=sys.stderr)
        print(
            "[sari] fix: ensure SARI_CONFIG points to a valid JSON file and db_path targets a .db file.",
            file=sys.stderr,
        )
        return 2


    # Security hardening: loopback-only by default.
    # Allow opt-in override only when explicitly requested.
    allow_non_loopback = os.environ.get("SARI_ALLOW_NON_LOOPBACK") == "1"
    host = _resolve_http_host(cfg.http_api_host, allow_non_loopback)

    # Unified DB path (single DB across workspaces)
    # DB path is determined by Config.load
    db_path = cfg.db_path
    Path(db_path).parent.mkdir(parents=True, exist_ok=True)

    print(f"[sari] DB path: {db_path}", file=sys.stderr)

    db = LocalSearchDB(db_path)
    db.set_settings(settings)
    try:
        from sari.core.engine_registry import get_default_engine
        db.set_engine(get_default_engine(db, cfg, cfg.workspace_roots))
    except Exception as e:
        print(f"[sari] engine init failed: {e}", file=sys.stderr)
    indexer = Indexer(cfg, db)
    # Wire search throttling coordinator to DB for read-priority hints
    db.coordinator = getattr(indexer, "coordinator", None)

    # Start HTTP immediately so health checks don't block on initial indexing.
    version = _resolve_version()
    mcp_server = _create_mcp_server(workspace_root, cfg, db, indexer)
    
    # Prioritize Async server
    try:
        from .async_http_server import serve_async
        httpd, actual_port = serve_async(
            host,
            cfg.http_api_port,
            db,
            indexer,
            version=version,
            workspace_root=workspace_root,
            cfg=cfg,
            mcp_server=mcp_server,
        )
    except Exception as e:
        print(f"[sari] async server start failed: {e}. falling back to sync.", file=sys.stderr)
        httpd, actual_port = serve_forever(
            host,
            cfg.http_api_port,
            db,
            indexer,
            version=version,
            workspace_root=workspace_root,
            cfg=cfg,
            mcp_server=mcp_server,
        )

    # Write server.json with actual binding info (single source of truth for port tracking)
    _write_server_info(workspace_root, host, actual_port, cfg.http_api_port)

    if actual_port != cfg.http_api_port:
        print(f"[sari] server.json updated with fallback port {actual_port}", file=sys.stderr)

    try:
        port_file = Path(db_path + ".http_api.port")
        port_file.write_text(str(actual_port) + "\n", encoding="utf-8")
    except Exception:
        pass

    stop_evt = threading.Event()

    def _shutdown(*_):
        if stop_evt.is_set():
            return
        stop_evt.set()
        try:
            indexer.stop()
        except Exception:
            pass
        try:
            if mcp_server and hasattr(mcp_server, "shutdown"):
                mcp_server.shutdown()
        except Exception:
            pass
        try:
            httpd.shutdown()
        except Exception:
            pass
        try:
            if hasattr(db, "close_all"):
                db.close_all()
            else:
                db.close()
        except Exception:
            pass

    signal.signal(signal.SIGINT, _shutdown)
    signal.signal(signal.SIGTERM, _shutdown)

    # Index in background.
    idx_thread = threading.Thread(target=indexer.run_forever, daemon=True)
    idx_thread.start()

    try:
        while not stop_evt.is_set():
            time.sleep(0.2)
    finally:
        _shutdown()

    return 0


if __name__ == "__main__":
    raise SystemExit(main())



from collections.abc import Mapping, Sequence
from typing import TypeAlias

TreeNode: TypeAlias = Mapping[str, object]


def render_tree(tree: TreeNode, max_print_depth: int = 2) -> str:
    """
    호출 그래프 트리를 ASCII 형식의 텍스트로 렌더링합니다.
    디버깅 및 사용자에게 직관적인 시각화를 제공합니다.
    
    Args:
        tree: 중첩된 딕셔너리 형태의 트리 구조
        max_print_depth: 보여줄 최대 깊이
        
    Returns:
        ASCII 트리 문자열
    """
    lines = []
    
    def _visit(node: TreeNode, prefix: str = "", is_last: bool = True, depth: int = 0):
        if depth > max_print_depth:
            lines.append(f"{prefix}...")
            return

        name = node.get("name") or "?"
        path = node.get("path") or ""
        
        # 파일 경로의 마지막 부분(basename)만 표시하여 간결하게 표현
        short_path = path.split("/")[-1] if path else ""
        
        rel = node.get("rel_type")
        rel_tag = f"[{rel}] " if rel else ""
        
        # 트리 구조 표현을 위한 가지(branch) 문자 선택 (├── 또는 └──)
        connector = "└── " if is_last else "├── "
        display = f"{prefix}{connector}{rel_tag}{name} ({short_path})" if depth > 0 else f"{name} ({short_path})"
        lines.append(display)

        raw_children = node.get("children", [])
        children: list[TreeNode] = []
        if isinstance(raw_children, Sequence) and not isinstance(raw_children, (str, bytes, bytearray)):
            children = [c for c in raw_children if isinstance(c, Mapping)]
        # 중요도(confidence) 순으로 정렬하여 상위 결과 우선 표시
        children.sort(key=lambda x: x.get("confidence", 0), reverse=True)
        
        count = len(children)
        for i, child in enumerate(children):
            # 다음 뎁스를 위한 들여쓰기 접두어 계산
            new_prefix = prefix + ("    " if is_last else "│   ")
            _visit(child, new_prefix, i == count - 1, depth + 1)

    _visit(tree)
    return "\n".join(lines)



import struct
from sari.mcp.tools.call_graph import build_call_graph
from sari.core.models import SearchOptions, IndexingResult


def _prepare_db_v2(conn):
    """테스트용 DB에 v2 스키마(importance_score) 강제 적용"""
    try:
        conn.execute(
            "ALTER TABLE symbols ADD COLUMN importance_score REAL DEFAULT 0.0")
    except Exception:
        pass


def test_intelligence_fuzzy_autorun(db, workspace):
    """오타가 포함된 심볼 조회 시 Fuzzy 알고리즘 작동 검증"""
    rid = workspace
    conn = db.get_read_connection()
    _prepare_db_v2(conn)

    res = IndexingResult(
        path=f"{rid}/user.py",
        rel="user.py",
        root_id=rid,
        repo="repo",
        type="new")
    db.upsert_files_turbo([res.to_file_row()])
    db.finalize_turbo_batch()

    conn.execute(
        "INSERT INTO symbols (symbol_id, path, root_id, name, kind, line, end_line, content, qualname) VALUES (?,?,?,?,?,?,?,?,?)",
        ("s1",
         f"{rid}/user.py",
         rid,
         "UserService",
         "class",
         1,
         10,
         "",
         "UserService"))

    graph = build_call_graph({"name": "UserServcie", "depth": 1}, db, [rid])
    assert graph["symbol"] == "UserService"
    assert "fuzzy match" in graph.get("scope_reason", "").lower()


def test_intelligence_centrality_ranking(db, workspace):
    """중요도(Centrality) 알고리즘이 랭킹에 미치는 영향 검증"""
    rid = workspace
    conn = db.get_read_connection()
    _prepare_db_v2(conn)

    # 1. 두 개의 파일 준비 (fts_content에 공통 키워드 'core' 주입)
    # File A: 중요도 높음
    row_a = IndexingResult(
        path=f"{rid}/engine.py",
        rel="engine.py",
        root_id=rid,
        repo="repo",
        type="new",
        fts_content="core engine logic").to_file_row()
    # File B: 중요도 낮음
    row_b = IndexingResult(
        path=f"{rid}/util.py",
        rel="util.py",
        root_id=rid,
        repo="repo",
        type="new",
        fts_content="core utility helper").to_file_row()

    db.upsert_files_turbo([row_a, row_b])
    db.finalize_turbo_batch()

    # 심볼 중요도 강제 설정
    conn.execute(
        "INSERT INTO symbols (symbol_id, path, root_id, name, kind, line, end_line, content, qualname, importance_score) VALUES (?,?,?,?,?,?,?,?,?,?)",
        ("s-eng",
         f"{rid}/engine.py",
         rid,
         "MainEngine",
         "class",
         1,
         10,
         "",
         "MainEngine",
         100.0))
    conn.execute(
        "INSERT INTO symbols (symbol_id, path, root_id, name, kind, line, end_line, content, qualname, importance_score) VALUES (?,?,?,?,?,?,?,?,?,?)",
        ("s-util",
         f"{rid}/util.py",
         rid,
         "Helper",
         "func",
         1,
         10,
         "",
         "Helper",
         1.0))

    # 2. 'core' 키워드로 검색 (두 파일 모두 매칭되지만 순위가 갈려야 함)
    opts = SearchOptions(query="core", root_ids=[rid])
    hits, _ = db.search_v2(opts)

    # 3. 검증
    assert len(hits) >= 2
    assert hits[0].path == f"{rid}/engine.py"  # 중요도 100점짜리가 1등이어야 함
    assert "importance=100.0" in hits[0].hit_reason


def test_intelligence_semantic_similarity(db, workspace):
    """벡터 유사도(Semantic)를 통한 의미 기반 검색 검증"""
    rid = workspace
    conn = db.get_read_connection()
    _prepare_db_v2(conn)

    vec = struct.pack("3f", 1.0, 0.0, 0.0)
    conn.execute(
        "INSERT INTO embeddings (root_id, entity_type, entity_id, content_hash, model, vector, created_ts, updated_ts) VALUES (?,?,?,?,?,?,?,?)",
        (rid,
         "file",
         f"{rid}/auth.py",
         "h1",
         "text-embedding-3",
         vec,
         0,
         0))

    query_vector = [0.9, 0.1, 0.0]
    hits = db.search_repo.search_semantic(
        query_vector, limit=5, root_ids=[rid])

    assert len(hits) > 0
    assert hits[0].path == f"{rid}/auth.py"
    assert hits[0].score > 80.0



import threading
import logging
import time
from typing import Dict
from sari.core.config.manager import ConfigManager
from sari.core.db.main import LocalSearchDB
from sari.core.indexer import Indexer
from sari.core.search_engine import SearchEngine
from sari.core.workspace import WorkspaceManager

logger = logging.getLogger("sari.registry")


def _resolve_workspace_roots_for_indexing(
        workspace_root: str, configured_roots: object) -> list[str]:
    """
    Resolve and sanitize roots for indexing.
    - Keep only existing directories.
    - Always include current workspace root if it exists.
    - Preserve order while deduplicating.
    """
    roots: list[str] = []
    raw: list[str] = []

    if isinstance(configured_roots, list):
        raw.extend(str(r) for r in configured_roots if r)
    raw.append(workspace_root)

    seen: set[str] = set()
    for entry in raw:
        norm = WorkspaceManager.normalize_path(entry)
        try:
            from pathlib import Path
            p = Path(norm).expanduser()
            resolved = WorkspaceManager.normalize_path(str(p))
            if not p.exists() or not p.is_dir():
                logger.warning("Skipping invalid workspace root: %s", resolved)
                continue
        except Exception:
            continue

        if resolved in seen:
            continue
        seen.add(resolved)
        roots.append(resolved)

    return roots


class SharedState:
    def __init__(self, workspace_root: str):
        self.workspace_root = WorkspaceManager.normalize_path(workspace_root)
        try:
            from sari.core.settings import settings
            WorkspaceManager.set_settings(settings)
        except Exception:
            pass
        self.root_id = WorkspaceManager.root_id_for_workspace(workspace_root)

        # 1. Load Layered Config (Phase 1)
        self.config_manager = ConfigManager(workspace_root)
        self.config_data = self.config_manager.resolve_final_config()

        # 2. Init DB (Phase 4) - Force global DB only
        db_path = str(WorkspaceManager.get_global_db_path())
        local_dir = WorkspaceManager.get_workspace_data_dir(workspace_root)
        if local_dir.exists():
            logger.info(
                f"Workspace-local DB detected; ignoring in favor of global DB: {db_path}")
        else:
            logger.info(f"Using global DB: {db_path}")

        self.db = LocalSearchDB(db_path)
        try:
            from sari.core.settings import settings
            self.db.set_settings(settings)
        except Exception:
            pass

        # 3. Init Indexer (Phase 3 & 2)
        from sari.core.config import Config
        defaults = Config.get_defaults(workspace_root)
        try:
            cfg_path = WorkspaceManager.resolve_config_path(workspace_root)
            loaded_cfg = Config.load(
                cfg_path,
                workspace_root_override=workspace_root)
            configured_roots = getattr(loaded_cfg, "workspace_roots", [])
        except Exception:
            configured_roots = [workspace_root]

        index_roots = _resolve_workspace_roots_for_indexing(
            workspace_root, configured_roots)
        defaults["workspace_roots"] = index_roots or [workspace_root]
        defaults["include_ext"] = self.config_data.get(
            "final_extensions", defaults["include_ext"])
        defaults["include_files"] = self.config_data.get(
            "final_filenames", defaults["include_files"])
        defaults["exclude_dirs"] = self.config_data.get(
            "final_exclude_dirs", defaults["exclude_dirs"])
        defaults["exclude_globs"] = self.config_data.get(
            "final_exclude_globs", defaults["exclude_globs"])
        defaults["gitignore_lines"] = self.config_data.get(
            "gitignore_lines", defaults.get("gitignore_lines", []))
        cfg_obj = Config(**defaults)

        self.indexer = Indexer(cfg_obj, self.db, logger=logger)

        # 4. Init File Watcher (Real-time sync)
        try:
            from sari.core.watcher import FileWatcher

            def on_change(evt):
                if self.indexer:
                    # evt.path is already the file path
                    self.indexer.index_file(evt.path)

            self.watcher = FileWatcher(
                list(cfg_obj.workspace_roots), on_change_callback=on_change, logger=logger)
        except Exception as e:
            logger.warning(f"Failed to initialize FileWatcher: {e}")
            self.watcher = None

        # 5. Wire Scheduling Coordinator for Read-Priority (Deprecated/Unused)
        # self.db.coordinator = self.indexer.coordinator

        try:
            from sari.core.engine_registry import get_default_engine
            self.db.set_engine(
                get_default_engine(
                    self.db,
                    cfg_obj,
                    cfg_obj.workspace_roots))
        except Exception:
            self.search_engine = SearchEngine(self.db)
            self.db.set_engine(self.search_engine)

        # 5. MCP Server instance (Lazy import to avoid circularity)
        from sari.mcp.server import LocalSearchMCPServer
        self.server = LocalSearchMCPServer(workspace_root)

        self.http_port = None
        self.http_host = None
        self.httpd = None
        self.last_activity = time.time()
        self.ref_count = 0
        self.persistent = False
        self._lock = threading.Lock()

    def start(self):
        # 0. Ensure all configured roots exist in DB metadata.
        root_paths = []
        try:
            root_paths = list(getattr(getattr(self.indexer, "cfg", None), "workspace_roots", []) or [])
        except Exception:
            root_paths = []
        if not root_paths:
            root_paths = [str(self.workspace_root)]

        for root_path in root_paths:
            try:
                normalized = WorkspaceManager.normalize_path(str(root_path))
                rid = WorkspaceManager.root_id_for_workspace(normalized)
                self.db.ensure_root(rid, normalized)
            except Exception as e:
                logger.error(f"Failed to ensure root for {root_path}: {e}")

        # 1. Start Indexer
        threading.Thread(target=self.indexer.run_forever, daemon=True).start()
        try:
            self.indexer.request_rescan()
        except Exception:
            pass

        # 1.1 Start Watcher
        if hasattr(self, "watcher") and self.watcher:
            try:
                self.watcher.start()
                logger.info(f"FileWatcher started for {self.workspace_root}")
            except Exception as e:
                logger.error(f"Failed to start FileWatcher: {e}")

    def touch(self):
        with self._lock:
            self.last_activity = time.time()

    def stop(self):
        try:
            if hasattr(self, "watcher") and self.watcher:
                self.watcher.stop()
            self.indexer.stop()
        except Exception as e:
            logger.error(f"Error stopping indexer/watcher: {e}")

        # 1. Cleanup Registry
        try:
            from sari.core.server_registry import ServerRegistry
            ServerRegistry().unregister_workspace(self.workspace_root)
        except Exception:
            pass

        # 2. Close DB
        try:
            self.db.close_all()
        except Exception:
            try:
                self.db.close()
            except Exception:
                pass


class Registry:
    _instance = None
    _lock = threading.Lock()
    def __init__(self): self._sessions: Dict[str, SharedState] = {}

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = cls()
        return cls._instance

    def get_or_create(
            self,
            workspace_root: str,
            persistent: bool = False,
            track_ref: bool = True) -> SharedState:
        with self._lock:
            if workspace_root not in self._sessions:
                session = SharedState(workspace_root)
                session.persistent = bool(persistent)
                session.start()
                self._sessions[workspace_root] = session
            elif persistent:
                self._sessions[workspace_root].persistent = True
            if track_ref:
                self._sessions[workspace_root].ref_count += 1
            self._sessions[workspace_root].touch()
            return self._sessions[workspace_root]

    def touch_workspace(self, workspace_root: str):
        with self._lock:
            if workspace_root in self._sessions:
                self._sessions[workspace_root].touch()

    def release(self, workspace_root: str):
        with self._lock:
            if workspace_root in self._sessions:
                state = self._sessions[workspace_root]
                state.ref_count = max(0, state.ref_count - 1)
                state.touch()
                if state.ref_count == 0 and not state.persistent:
                    state.stop()
                    del self._sessions[workspace_root]

    def shutdown_all(self):
        with self._lock:
            for s in self._sessions.values():
                s.stop()
            self._sessions.clear()

    def active_count(self) -> int:
        with self._lock:
            return sum(1 for s in self._sessions.values() if s.ref_count > 0)

    def has_persistent(self) -> bool:
        with self._lock:
            return any(s.persistent for s in self._sessions.values())

    def get_last_activity_ts(self) -> float:
        with self._lock:
            return max(
                (s.last_activity for s in self._sessions.values() if s.ref_count > 0),
                default=0.0)



"""
Tests for logging utilities module.
"""

import pytest
from sari.core.logging_utils import (
    safe_log,
    LoggerMixin,
    create_error_context,
    LogContext,
)


class MockLogger:
    """Mock logger for testing."""
    
    def __init__(self):
        self.messages = []
    
    def log_info(self, msg):
        self.messages.append(("info", msg))
    
    def log_error(self, msg):
        self.messages.append(("error", msg))
    
    def log_warning(self, msg):
        self.messages.append(("warning", msg))
    
    def log_debug(self, msg):
        self.messages.append(("debug", msg))


class StdlibStyleLogger:
    """Mock stdlib-style logger."""
    
    def __init__(self):
        self.messages = []
    
    def info(self, msg):
        self.messages.append(("info", msg))
    
    def error(self, msg):
        self.messages.append(("error", msg))
    
    def warning(self, msg):
        self.messages.append(("warning", msg))
    
    def debug(self, msg):
        self.messages.append(("debug", msg))


def test_safe_log_with_telemetry_logger():
    """Test safe_log with TelemetryLogger-style logger."""
    logger = MockLogger()
    
    safe_log(logger, "info", "Test info message")
    safe_log(logger, "error", "Test error message")
    safe_log(logger, "warning", "Test warning message")
    
    assert len(logger.messages) == 3
    assert logger.messages[0] == ("info", "Test info message")
    assert logger.messages[1] == ("error", "Test error message")
    assert logger.messages[2] == ("warning", "Test warning message")


def test_safe_log_with_stdlib_logger():
    """Test safe_log with stdlib-style logger."""
    logger = StdlibStyleLogger()
    
    safe_log(logger, "info", "Test info")
    safe_log(logger, "error", "Test error")
    
    assert len(logger.messages) == 2
    assert logger.messages[0] == ("info", "Test info")
    assert logger.messages[1] == ("error", "Test error")


def test_safe_log_with_none_logger():
    """Test safe_log with None logger (should not crash)."""
    result = safe_log(None, "info", "This should be silently ignored")
    assert result is None


def test_logger_mixin():
    """Test LoggerMixin class."""
    
    class TestClass(LoggerMixin):
        def __init__(self, logger):
            self.logger = logger
    
    logger = MockLogger()
    obj = TestClass(logger)
    
    obj.log_info("Info message")
    obj.log_error("Error message")
    obj.log_warning("Warning message")
    obj.log_debug("Debug message")
    
    assert len(logger.messages) == 4
    assert logger.messages[0] == ("info", "Info message")
    assert logger.messages[1] == ("error", "Error message")
    assert logger.messages[2] == ("warning", "Warning message")
    assert logger.messages[3] == ("debug", "Debug message")


def test_logger_mixin_without_logger():
    """Test LoggerMixin without a logger (should not crash)."""
    
    class TestClass(LoggerMixin):
        pass
    
    obj = TestClass()
    result = obj.log_info("This should be silently ignored")
    assert result is None


def test_create_error_context():
    """Test create_error_context function."""
    error = ValueError("Invalid input")
    
    # Without context
    msg = create_error_context("processing", error)
    assert msg == "processing failed: Invalid input"
    
    # With context
    msg = create_error_context("processing", error, file="test.txt", line=42)
    assert "processing failed: Invalid input" in msg
    assert "file=test.txt" in msg
    assert "line=42" in msg


def test_log_context_success():
    """Test LogContext for successful operation."""
    logger = MockLogger()
    
    with LogContext(logger, "test operation", key="value"):
        pass
    
    assert len(logger.messages) == 2
    assert logger.messages[0] == ("info", "test operation started (key=value)")
    assert logger.messages[1] == ("info", "test operation completed (key=value)")


def test_log_context_with_error():
    """Test LogContext when an error occurs."""
    logger = MockLogger()
    
    with pytest.raises(ValueError):
        with LogContext(logger, "test operation", key="value"):
            raise ValueError("Test error")
    
    assert len(logger.messages) == 2
    assert logger.messages[0] == ("info", "test operation started (key=value)")
    assert "test operation failed: Test error" in logger.messages[1][1]
    assert "key=value" in logger.messages[1][1]


def test_log_context_no_start_log():
    """Test LogContext with log_start=False."""
    logger = MockLogger()
    
    with LogContext(logger, "test operation", log_start=False):
        pass
    
    assert len(logger.messages) == 1
    assert logger.messages[0][0] == "info"
    assert "completed" in logger.messages[0][1]


def test_log_context_no_end_log():
    """Test LogContext with log_end=False."""
    logger = MockLogger()
    
    with LogContext(logger, "test operation", log_end=False):
        pass
    
    assert len(logger.messages) == 1
    assert logger.messages[0][0] == "info"
    assert "started" in logger.messages[0][1]


def test_log_context_with_none_logger():
    """Test LogContext with None logger (should not crash)."""
    with LogContext(None, "test operation"):
        marker = True
    assert marker is True



import re
import hashlib
from typing import List
from sari.core.models import ParserSymbol

def _symbol_id(path: str, kind: str, name: str) -> str:
    h = hashlib.sha1(f"{path}:{kind}:{name}".encode()).hexdigest()
    return h

class SpecialParser:
    """
    AST 분석이 어렵거나 구조가 단순한 특수 형식 파일(Dockerfile, MyBatis, Markdown 등)을 위한 전담 파서 집합입니다.
    """
    @staticmethod
    def parse_dockerfile(path: str, content: str) -> List[ParserSymbol]:
        """Dockerfile 지시어(FROM, RUN 등)를 심볼로 추출합니다."""
        symbols = []
        for i, line in enumerate(content.splitlines()):
            raw = line.strip()
            if not raw or raw.startswith("#"):
                continue
            m = re.match(r"^([A-Z]+)\b", raw)
            if not m:
                continue
            instr = m.group(1)
            sid = _symbol_id(path, "instruction", instr)
            symbols.append(ParserSymbol(
                sid=sid, path=path, name=instr, kind="instruction",
                line=i + 1, end_line=i + 1, content=raw,
                meta={"instruction": instr}, qualname=instr
            ))
        return symbols

    @staticmethod
    def parse_mybatis(path: str, content: str) -> List[ParserSymbol]:
        """MyBatis XML 파일에서 SQL 매핑 구문(select, insert 등)의 ID를 심볼로 추출합니다."""
        symbols = []
        for i, line in enumerate(content.splitlines()):
            m = re.search(r'<(select|insert|update|delete|sql)\s+id=["\']([^"\']+)["\']', line)
            if m:
                tag, name = m.group(1), m.group(2)
                sid = _symbol_id(path, "method", name)
                symbols.append(ParserSymbol(
                    sid=sid, path=path, name=name, kind="method",
                    line=i + 1, end_line=i + 1, content=line.strip(),
                    meta={"mybatis_tag": tag, "framework": "MyBatis"}, qualname=name
                ))
        return symbols

    @staticmethod
    def parse_markdown(path: str, content: str) -> List[ParserSymbol]:
        """Markdown 파일의 헤더(#, ## 등)를 구조적 심볼로 추출합니다."""
        symbols = []
        for i, line in enumerate(content.splitlines()):
            m = re.match(r"^(#+)\s+(.*)", line.strip())
            if m:
                lvl, name = len(m.group(1)), m.group(2)
                sid = _symbol_id(path, "doc", name)
                symbols.append(ParserSymbol(
                    sid=sid, path=path, name=name, kind="doc",
                    line=i + 1, end_line=i + 1, content=line.strip(),
                    meta={"lvl": lvl}, qualname=name
                ))
        return symbols



import pytest
import os
from sari.core.indexer.main import Indexer
from sari.core.db.main import LocalSearchDB
from sari.core.workspace import WorkspaceManager
from sari.core.config import Config

@pytest.fixture
def func_env(tmp_path):
    ws = (tmp_path / "func_ws").resolve()
    ws.mkdir()
    db = LocalSearchDB(str(tmp_path / "func.db"))
    # Ensure Config is based on the resolved absolute path
    cfg = Config(**Config.get_defaults(str(ws)))
    return {"ws": ws, "db": db, "cfg": cfg}

def test_symbol_extraction_integrity(func_env):
    """
    1. Verify that parallel workers correctly extract and store code symbols.
    """
    ws, db, cfg = func_env["ws"], func_env["db"], func_env["cfg"]
    
    # Use a raw string with precise indentation to avoid parsing issues
    code = (
        "class MyTruth:\n"
        "    def verify(self):\n"
        "        return True\n"
    )
    target_file = ws / "logic.py"
    target_file.write_text(code, encoding="utf-8")
    
    indexer = Indexer(cfg, db)
    indexer.scan_once()
    
    # Truth: Query the symbols table directly using broad match
    conn = db.db.connection()
    # We use LIKE to avoid absolute/relative path mismatch issues in tests
    res = conn.execute("SELECT name, kind FROM symbols WHERE name = 'MyTruth'").fetchone()
    
    assert res is not None, "Class symbol 'MyTruth' should be in DB"
    assert "class" in res[1].lower()
    
    res = conn.execute("SELECT name FROM symbols WHERE name = 'verify'").fetchone()
    assert res is not None, "Function symbol 'verify' should be in DB"

def test_stale_data_pruning_integrity(func_env):
    """
    2. Verify that deleted files are correctly removed from the DB.
    """
    ws, db, cfg = func_env["ws"], func_env["db"], func_env["cfg"]
    
    p1 = ws / "exists.py"
    p1.write_text("content")
    p2 = ws / "gone.py"
    p2.write_text("content")
    
    indexer = Indexer(cfg, db)
    indexer.scan_once()
    
    # Verify both exist
    assert len(db.search_files("exists.py")) == 1
    assert len(db.search_files("gone.py")) == 1
    
    # DELETE GONE.PY
    os.remove(p2)
    
    # RE-SCAN: Finds only p1
    indexer.scan_once()
    
    # ACTIVATE PRUNING: Only p1 remains active
    active_paths = [str(p1)]
    rid = WorkspaceManager.root_id(str(ws))
    db.prune_stale_data(rid, active_paths)
    
    # VERIFY: GONE.PY is wiped, EXISTS.PY remains
    assert len(db.search_files("gone.py")) == 0
    assert len(db.search_files("exists.py")) == 1

def test_search_hit_metadata_integrity(func_env):
    """
    3. Verify that search results return accurate metadata from Turbo DB.
    """
    ws, db, cfg = func_env["ws"], func_env["db"], func_env["cfg"]
    
    (ws / "meta.py").write_text("search_me_now")
    st = (ws / "meta.py").stat()
    
    indexer = Indexer(cfg, db)
    indexer.scan_once()
    
    results = db.search_files("search_me_now")
    assert len(results) == 1
    hit = results[0]
    
    # Metadata Truth: Path must be correct and size must match
    assert "meta.py" in hit["path"]
    assert hit["size"] == st.st_size


from sari.core.parsers.factory import ParserFactory

def test_python_parser_pure_logic():
    """
    Verify if the Python parser actually works in the current environment.
    """
    code = "class MyTruth:\n    def verify(self):\n        return True\n"
    parser = ParserFactory.get_parser(".py")
    
    assert parser is not None, "ParserFactory must return a Python parser"
    
    symbols, _ = parser.extract("test.py", code)
    
    print(f"\nDEBUG: Extracted symbols: {symbols}")
    
    names = [s.name for s in symbols]
    assert "MyTruth" in names
    assert "verify" in names


import re
from typing import List, Dict
from pathlib import Path
from .base import BaseParser
from .common import _symbol_id, _safe_compile, NORMALIZE_KIND_BY_EXT
from sari.core.models import ParserSymbol, ParserRelation, ParseResult


class GenericRegexParser(BaseParser):
    """
    Tree-sitter 라이브러리가 없거나 해당 언어를 지원하지 않을 때 사용하는 범용 정규식 파서입니다.
    미리 정의된 정규식 패턴을 사용하여 클래스와 메서드 구조를 휴리스틱하게 추출합니다.
    """

    def __init__(self, config: Dict[str, object], ext: str):
        self.ext = ext.lower()
        self.re_class = config["re_class"]
        self.re_method = config["re_method"]
        self.method_kind = config.get("method_kind", "method")
        self.re_anno = _safe_compile(r"^\s*@([a-zA-Z0-9_]+)")
        self.kind_norm = NORMALIZE_KIND_BY_EXT.get(self.ext, {})

    def sanitize(self, line: str) -> str:
        line = re.sub(r"//.*$", "", line)
        line = re.sub(r"#.*$", "", line)
        return line

    def extract(self,
                path: str,
                content: str) -> ParseResult:
        symbols: List[ParserSymbol] = []
        relations: List[ParserRelation] = []
        # Strip block comments first to avoid finding fake symbols in them
        content = re.sub(r"/\*.*?\*/", "", content, flags=re.DOTALL)
        lines = content.splitlines()
        active_scopes = []
        cur_bal = 0

        for i, line in enumerate(lines):
            line_no = i + 1
            clean = self.sanitize(line)
            if not clean.strip():
                continue

            matches = []
            for m in self.re_class.finditer(clean):
                name = m.group(2) if len(m.groups()) >= 2 else m.group(1)
                kind_raw = m.group(1).lower().strip()
                kind = self.kind_norm.get(kind_raw, "class")
                matches.append((name, kind, m.start()))

            for m in self.re_method.finditer(clean):
                name = next((g for g in m.groups() if g), None)
                if name and not any(name == x[0] for x in matches):
                    matches.append((name, self.method_kind, m.start()))

            for name, kind, _ in matches:
                sid = _symbol_id(path, kind, name)
                info = {
                    "sid": sid,
                    "path": path,
                    "name": name,
                    "kind": kind,
                    "line": line_no,
                    "meta": {},
                    "raw": line.strip(),
                    "qual": name}
                active_scopes.append((cur_bal, info))

            # Safer brace counting
            tmp_line = re.sub(r'"[^"\\]*(?:\\.[^"\\]*)*"', "", clean)
            tmp_line = re.sub(r"'[^'\\]*(?:\\.[^'\\]*)*'", "", tmp_line)
            op, cl = tmp_line.count("{"), tmp_line.count("}")
            cur_bal += (op - cl)

            still_active = []
            for bal, info in active_scopes:
                if cur_bal <= bal:
                    symbols.append(
                        ParserSymbol(
                            sid=info["sid"],
                            path=info["path"],
                            name=info["name"],
                            kind=info["kind"],
                            line=info["line"],
                            end_line=line_no,
                            content=info["raw"],
                            meta=info["meta"],
                            qualname=info["qual"]))
                else:
                    still_active.append((bal, info))
            active_scopes = still_active

        for _, info in active_scopes:
            symbols.append(
                ParserSymbol(
                    sid=info["sid"],
                    path=info["path"],
                    name=info["name"],
                    kind=info["kind"],
                    line=info["line"],
                    end_line=len(lines),
                    content=info["raw"],
                    meta=info["meta"],
                    qualname=info["qual"]))

        if self.ext == ".vue":
            stem = Path(path).stem
            symbols.append(
                ParserSymbol(
                    sid=_symbol_id(
                        path,
                        "class",
                        stem),
                    path=path,
                    name=stem,
                    kind="class",
                    line=1,
                    end_line=len(lines),
                    content=stem,
                    qualname=stem))

        return ParseResult(symbols=symbols, relations=relations)


class HCLRegexParser(GenericRegexParser):
    """
    Terraform(HCL) 파일을 처리하기 위한 특화된 정규식 파서입니다.
    resource, module, variable 등 HCL 특유의 블록 구조를 처리합니다.
    """

    def __init__(self, config: Dict[str, object], ext: str):
        super().__init__(config, ext)
        self.re_resource = _safe_compile(r'^resource\s+"([^"]+)"\s+"([^"]+)"')
        self.re_general = _safe_compile(
            r'^(module|variable|output|data)\s+"([^"]+)"')

    def extract(self,
                path: str,
                content: str) -> ParseResult:
        symbols: List[ParserSymbol] = []
        relations: List[ParserRelation] = []
        content = re.sub(r"/\*.*?\*/", "", content, flags=re.DOTALL)
        lines = content.splitlines()
        active_scopes = []
        cur_bal = 0

        for i, line in enumerate(lines):
            line_no = i + 1
            clean = self.sanitize(line)
            if not clean.strip():
                continue

            matches = []
            for m in self.re_resource.finditer(clean):
                rtype, rname = m.group(1), m.group(2)
                name = f"{rtype}.{rname}"
                matches.append((name, "resource", m.start()))

            if not matches:
                for m in self.re_general.finditer(clean):
                    btype, bname = m.group(1), m.group(2)
                    matches.append((bname, btype, m.start()))

            for name, kind, _ in matches:
                sid = _symbol_id(path, kind, name)
                info = {
                    "sid": sid,
                    "path": path,
                    "name": name,
                    "kind": kind,
                    "line": line_no,
                    "meta": {},
                    "raw": line.strip(),
                    "qual": name}
                active_scopes.append((cur_bal, info))

            tmp_line = re.sub(r'"[^"\\]*(?:\\.[^"\\]*)*"', "", clean)
            tmp_line = re.sub(r"'[^'\\]*(?:\\.[^'\\]*)*'", "", tmp_line)
            op, cl = tmp_line.count("{"), tmp_line.count("}")
            cur_bal += (op - cl)

            still_active = []
            for bal, info in active_scopes:
                if cur_bal <= bal:
                    symbols.append(
                        ParserSymbol(
                            sid=info["sid"],
                            path=info["path"],
                            name=info["name"],
                            kind=info["kind"],
                            line=info["line"],
                            end_line=line_no,
                            content=info["raw"],
                            meta=info["meta"],
                            qualname=info["qual"]))
                else:
                    still_active.append((bal, info))
            active_scopes = still_active

        for _, info in active_scopes:
            symbols.append(
                ParserSymbol(
                    sid=info["sid"],
                    path=info["path"],
                    name=info["name"],
                    kind=info["kind"],
                    line=info["line"],
                    end_line=len(lines),
                    content=info["raw"],
                    meta=info["meta"],
                    qualname=info["qual"]))

        return ParseResult(symbols=symbols, relations=relations)



from sari.mcp.middleware import run_middlewares


def test_run_middlewares_returns_error_on_non_object_result():
    res = run_middlewares("tool", None, {}, [], lambda: "bad-result")
    assert res.get("isError") is True
    assert res["error"]["code"] == -32000



import time
from typing import TypeAlias
from sari.mcp.tools._util import mcp_response, pack_error, ErrorCode

ToolResult: TypeAlias = dict[str, object]
UsageState: TypeAlias = dict[str, object]


class PolicyEngine:
    def __init__(self, mode: str = "warn"):
        self.mode = mode
        self.usage: UsageState = {
            "search": 0,
            "search_symbols": 0,
            "last_search_ts": None,
            "last_search_symbols_ts": None,
            "read_without_search": 0,
        }

    def mark_action(self, tool_name: str):
        now = time.time()
        if tool_name in {"search", "grep_and_read"}:
            self.usage["search"] += 1
            self.usage["last_search_ts"] = now
        elif tool_name == "search_symbols":
            self.usage["search_symbols"] += 1
            self.usage["last_search_symbols_ts"] = now

    def has_search_context(self) -> bool:
        search_count = int(self.usage.get("search", 0) or 0)
        symbol_search_count = int(self.usage.get("search_symbols", 0) or 0)
        return search_count > 0 or symbol_search_count > 0

    def check_pre_call(self, tool_name: str) -> ToolResult | None:
        if self.mode == "off":
            return None
            
        if tool_name in {"read_file", "read_symbol"}:
            if not self.has_search_context():
                self.usage["read_without_search"] += 1
                if self.mode == "enforce":
                    return mcp_response(
                        "search_first",
                        lambda: pack_error("search_first", ErrorCode.INVALID_ARGS, "search-first policy active. Call search/search_symbols before read_file/read_symbol."),
                        lambda: {"error": {"code": ErrorCode.INVALID_ARGS.value, "message": "search-first policy active. Call search/search_symbols before read_file/read_symbol."}, "isError": True},
                    )
        return None

    def apply_post_call(self, tool_name: str, result: ToolResult) -> ToolResult:
        if self.mode == "warn" and tool_name in {"read_file", "read_symbol"}:
            if not self.has_search_context():
                raw_warnings = result.get("warnings", [])
                warnings = list(raw_warnings) if isinstance(raw_warnings, list) else []
                if "Search-first policy (advisory): call search/search_symbols before read_file/read_symbol." not in warnings:
                    warnings.append("Search-first policy (advisory): call search/search_symbols before read_file/read_symbol.")
                result["warnings"] = warnings
        return result



import json
from collections.abc import Mapping
from typing import TypeAlias

from sari.mcp.tools._util import (
    mcp_response,
    pack_header,
    pack_line,
    pack_encode_id,
    pack_encode_text,
    resolve_root_ids,
    pack_error,
    ErrorCode,
    invalid_args_response,
)

ToolResult: TypeAlias = dict[str, object]

def execute_search_api_endpoints(args: object, db: object, roots: list[str]) -> ToolResult:
    """
    URL 경로 패턴을 기반으로 관련 API 엔드포인트(함수, 메서드)를 검색하는 도구입니다.
    코드 내 메타데이터에 기록된 `http_path` 정보를 활용하여 탐색합니다.
    """
    if not isinstance(args, Mapping):
        return invalid_args_response("search_api_endpoints", "args must be an object")

    path_query = str(args.get("path", "")).strip()
    if not path_query:
        return mcp_response(
            "search_api_endpoints",
            lambda: pack_error("search_api_endpoints", ErrorCode.INVALID_ARGS, "Path query is required"),
            lambda: {"error": {"code": ErrorCode.INVALID_ARGS.value, "message": "Path query is required"}, "isError": True},
        )

    repo = args.get("repo")

    # 메타데이터에 경로가 포함된 심볼 검색
    # SQLite JSON 지원 버전의 한계로 인해, 메타데이터 텍스트에 대해 LIKE 검색을 수행한 후 Python에서 필터링합니다.
    sql = """
        SELECT s.path, s.name, s.kind, s.line, s.meta_json AS metadata, s.content, f.repo
        FROM symbols s
        JOIN files f ON s.path = f.path
        WHERE s.meta_json LIKE ? AND (s.kind = 'method' OR s.kind = 'function' OR s.kind = 'class')
    """
    # 메타데이터에서 부분 일치 검색 (느슨한 LIKE 검색)
    params = [f'%{path_query}%']
    root_ids = resolve_root_ids(roots)
    if root_ids:
        root_clause = " OR ".join(["s.path LIKE ?"] * len(root_ids))
        sql += f" AND ({root_clause})"
        params.extend([f"{rid}/%" for rid in root_ids])
    if repo:
        sql += " AND f.repo = ?"
        params.append(repo)

    conn = db.get_read_connection() if hasattr(db, "get_read_connection") else db._read
    rows = conn.execute(sql, params).fetchall()

    results = []
    for r in rows:
        try:
            meta = json.loads(r["metadata"] or "{}")
            http_path = meta.get("http_path", "")
            # 쿼리가 http_path에 포함되는지 확인
            if path_query in http_path or path_query == http_path:
                results.append({
                    "path": r["path"],
                    "name": r["name"],
                    "kind": r["kind"],
                    "line": r["line"],
                    "repo": r["repo"],
                    "http_path": http_path,
                    "annotations": meta.get("annotations", []),
                    "snippet": r["content"]
                })
        except (json.JSONDecodeError, KeyError) as e:
            import logging
            logging.getLogger("sari.mcp.tools").debug(f"Failed to parse metadata for {r['path']}: {e}")
            continue

    def build_pack() -> str:
        """PACK1 형식의 응답을 생성합니다."""
        lines = [pack_header("search_api_endpoints", {"q": pack_encode_text(path_query)}, returned=len(results))]
        if not repo:
            lines.append(pack_line("m", {"hint": pack_encode_text("Narrow scope using repo or root_ids")}))
        for r in results:
            kv = {
                "path": pack_encode_id(r["path"]),
                "name": pack_encode_id(r["name"]),
                "kind": pack_encode_id(r["kind"]),
                "line": str(r["line"]),
                "http_path": pack_encode_text(r["http_path"]),
                "repo": pack_encode_id(r.get("repo", "")),
            }
            lines.append(pack_line("r", kv))
        return "\n".join(lines)

    return mcp_response(
        "search_api_endpoints",
        build_pack,
        lambda: {"query": path_query, "repo": repo or "", "results": results, "count": len(results), "meta": {"hint": "Narrow scope using repo or root_ids" if not repo else ""}},
    )



from collections.abc import Mapping
from typing import TypeAlias
import time

ConfigMap: TypeAlias = dict[str, object]


class ScoringPolicy:
    @staticmethod
    def _section(config: ConfigMap, key: str, defaults: dict[str, float]) -> dict[str, float]:
        raw = config.get(key)
        if isinstance(raw, Mapping):
            out: dict[str, float] = dict(defaults)
            for k, v in raw.items():
                try:
                    out[str(k)] = float(v)
                except Exception:
                    continue
            return out
        return dict(defaults)

    def __init__(self, config: object = None):
        self.config: ConfigMap = dict(config) if isinstance(config, Mapping) else {}
        
        # Symbol weights
        self.symbol_weights = self._section(self.config, "symbol_weights", {
            "class": 600.0,
            "function": 500.0,
            "method": 350.0,
            "interface": 450.0,
            "exact_match_bonus": 800.0,
            "symbol_priority_base": 1000.0
        })
        
        # File/Path weights
        self.path_weights = self._section(self.config, "path_weights", {
            "exact_filename": 2.0,
            "filename_stem": 1.2,
            "path_suffix": 1.0,
            "core_file_bonus": 0.2
        })
        
        # SQL based priors (from search_engine.py)
        self.sql_priors = self._section(self.config, "sql_priors", {
            "src_path": 0.6,
            "config_path": 0.4,
            "test_path": -0.7,
            "code_ext": 0.3,
            "config_ext": 0.15,
            "noise_ext": -0.8
        })

    def get_symbol_boost(self, kind: str, is_exact: bool = False) -> float:
        boost = self.symbol_weights.get(kind.lower(), 0.0)
        if is_exact:
            boost += self.symbol_weights.get("exact_match_bonus", 0.0)
        return self.symbol_weights.get("symbol_priority_base", 1000.0) + boost

    def calculate_recency_boost(self, mtime: int, base_score: float) -> float:
        now = time.time()
        age_days = (now - mtime) / 86400
        if age_days < 1:
            boost = 1.5
        elif age_days < 7:
            boost = 1.3
        elif age_days < 30:
            boost = 1.1
        else:
            boost = 1.0
        return (base_score + 0.1) * boost

    def get_path_prior_sql(self) -> str:
        p = self.sql_priors
        return f"""
        CASE
            WHEN f.path LIKE 'src/%' OR f.path LIKE '%/src/%' OR f.path LIKE 'app/%' OR f.path LIKE '%/app/%' OR f.path LIKE 'core/%' OR f.path LIKE '%/core/%' THEN {float(p['src_path'])}
            WHEN f.path LIKE 'config/%' OR f.path LIKE '%/config/%' OR f.path LIKE 'domain/%' OR f.path LIKE '%/domain/%' OR f.path LIKE 'service/%' OR f.path LIKE '%/service/%' THEN {float(p['config_path'])}
            WHEN f.path LIKE 'test/%' OR f.path LIKE '%/test/%' OR f.path LIKE 'tests/%' OR f.path LIKE '%/tests/%' OR f.path LIKE 'example/%' OR f.path LIKE '%/example/%' OR f.path LIKE 'dist/%' OR f.path LIKE '%/dist/%' OR f.path LIKE 'build/%' OR f.path LIKE '%/build/%' THEN {float(p['test_path'])}
            ELSE 0.0
        END
        """

    def get_filetype_prior_sql(self) -> str:
        p = self.sql_priors
        return f"""
        CASE
            WHEN f.path LIKE '%.py' OR f.path LIKE '%.ts' OR f.path LIKE '%.go' OR f.path LIKE '%.java' OR f.path LIKE '%.kt' THEN {float(p['code_ext'])}
            WHEN f.path LIKE '%.yaml' OR f.path LIKE '%.yml' OR f.path LIKE '%.json' THEN {float(p['config_ext'])}
            WHEN f.path LIKE '%.lock' OR f.path LIKE '%.min.js' OR f.path LIKE '%.map' THEN {float(p['noise_ext'])}
            ELSE 0.0
        END
        """



from sari.mcp.cli.smart_daemon import ensure_smart_daemon
from sari.mcp.trace import trace
from sari.core.daemon_resolver import resolve_daemon_address as _resolve_daemon_target
from sari.core.workspace import WorkspaceManager
from sari.mcp.telemetry import TelemetryLogger
import sys
import json
import socket
import threading
import os
import time
import logging
import secrets
from pathlib import Path

# Add project root to sys.path for absolute imports
SCRIPT_DIR = Path(__file__).parent
REPO_ROOT = SCRIPT_DIR.parent
if str(REPO_ROOT) not in sys.path:
    sys.path.insert(0, str(REPO_ROOT))


try:
    import fcntl  # type: ignore
except Exception:
    fcntl = None

# Configure logging to stderr so it doesn't interfere with MCP STDIO
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
    stream=sys.stderr
)
logger = logging.getLogger("mcp-proxy")
telemetry = TelemetryLogger(WorkspaceManager.get_global_log_dir())


def _log_info(message: str) -> None:
    logger.info(message)
    try:
        telemetry.log_info(message)
    except Exception:
        pass


def _log_error(message: str) -> None:
    logger.error(message)
    try:
        telemetry.log_error(message)
    except Exception:
        pass


DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = 47779
MAX_MESSAGE_SIZE = 10 * 1024 * 1024  # 10MB
_HEADER_SEP = b"\r\n\r\n"
_MODE_FRAMED = "framed"
_MODE_JSONL = "jsonl"


def _identify_sari_daemon(host: str, port: int, timeout: float = 0.3) -> bool:
    try:
        trace("proxy_identify_start", host=host, port=port)
        with socket.create_connection((host, port), timeout=timeout) as sock:
            sock.settimeout(timeout)
            body = json.dumps({"jsonrpc": "2.0", "id": 1,
                              "method": "sari/identify"}).encode("utf-8")
            header = f"Content-Length: {len(body)}\r\n\r\n".encode("ascii")
            sock.sendall(header + body)

            f = sock.makefile("rb")
            headers = {}
            while True:
                line = f.readline()
                if not line:
                    return False
                line = line.strip()
                if not line:
                    break
                if b":" in line:
                    k, v = line.split(b":", 1)
                    headers[k.strip().lower()] = v.strip()

            try:
                content_length = int(headers.get(b"content-length", b"0"))
            except ValueError:
                return False
            if content_length <= 0:
                return False
            resp_body = f.read(content_length)
            if not resp_body:
                return False
            resp = json.loads(resp_body.decode("utf-8"))
            result = resp.get("result") or {}
            if result.get("name") == "sari":
                trace("proxy_identify_ok", host=host, port=port)
                return True
    except Exception:
        trace("proxy_identify_error", host=host, port=port)
        pass
    return False


def _lock_file(lock_file) -> None:
    if fcntl is not None:
        fcntl.flock(lock_file, fcntl.LOCK_EX)
        return
    try:
        import msvcrt
        lock_file.seek(0)
        msvcrt.locking(lock_file.fileno(), msvcrt.LK_LOCK, 1)
    except Exception:
        pass


def _unlock_file(lock_file) -> None:
    if fcntl is not None:
        fcntl.flock(lock_file, fcntl.LOCK_UN)
        return
    try:
        import msvcrt
        lock_file.seek(0)
        msvcrt.locking(lock_file.fileno(), msvcrt.LK_UNLCK, 1)
    except Exception:
        pass


def start_daemon_if_needed(host, port, workspace_root: str = ""):
    """Ensures daemon is running using smart logic."""
    try:
        ensure_smart_daemon(host, port, workspace_root)
        return True
    except Exception as e:
        _log_error(f"Failed to start daemon at {host}:{port}: {e}")
        trace("proxy_start_daemon_error", host=host, port=port, error=str(e))
        return False


def forward_socket_to_stdout(sock, state):
    try:
        f = sock.makefile("rb")
        while True:
            # Read Headers
            headers = {}
            while True:
                line = f.readline()
                if not line:
                    break
                line_str = line.decode("utf-8").strip()
                if not line_str:
                    break
                if ":" in line_str:
                    k, v = line_str.split(":", 1)
                    headers[k.strip().lower()] = v.strip()

            if not headers and not line:
                break

            content_length = int(headers.get("content-length", 0))
            if content_length <= 0:
                continue

            body = f.read(content_length)
            if not body:
                break

            msg_id = None
            obj = None
            try:
                obj = json.loads(body.decode("utf-8"))
                if isinstance(obj, dict):
                    msg_id = obj.get("id")
            except Exception:
                msg_id = None
                obj = None

            # If daemon is draining, reconnect and replay initialize without
            # surfacing an error to the client.
            try:
                if isinstance(obj, dict):
                    err = obj.get("error") or {}
                    code = err.get("code")
                    message = str(err.get("message") or "")
                    if code == -32001 and "draining" in message.lower():
                        _log_info(
                            "Server draining detected; reconnecting to latest daemon.")
                        state["dead"] = True
                        if _reconnect(state):
                            init_req = state.get("init_request")
                            if init_req:
                                _send_payload(
                                    state,
                                    json.dumps(init_req).encode("utf-8"),
                                    state.get("mode") or _MODE_FRAMED,
                                )
                            continue
            except Exception:
                pass
            if msg_id is not None:
                with state["suppress_lock"]:
                    suppress_ids = state.setdefault("suppress_ids", set())
                    if msg_id in suppress_ids:
                        suppress_ids.discard(msg_id)
                        continue

            mode = state.get("mode") or _MODE_FRAMED
            if mode == _MODE_JSONL:
                sys.stdout.buffer.write(body + b"\n")
                sys.stdout.buffer.flush()
            else:
                header = f"Content-Length: {len(body)}\r\n\r\n".encode("ascii")
                sys.stdout.buffer.write(header + body)
                sys.stdout.buffer.flush()
            trace("proxy_forwarded_to_stdout", bytes=len(body))
    except Exception as e:
        _log_error(f"Error forwarding socket to stdout: {e}")
        trace("proxy_forward_stdout_error", error=str(e))
    finally:
        state["dead"] = True
        try:
            sock.close()
        except Exception:
            pass


def _read_mcp_message(stdin):
    """Robustly reads one MCP message, skipping noise or leading logs."""
    while True:
        line = stdin.readline()
        if not line:
            return None

        stripped = line.strip()
        if not stripped:
            continue

        # 1. JSONL fallback (starts with '{')
        if stripped.startswith(b"{"):
            return stripped, _MODE_JSONL

        # 2. Content-Length framing
        if stripped.lower().startswith(b"content-length:"):
            try:
                # Re-parse headers fully
                content_length = None
                headers = [line]
                while True:
                    h = stdin.readline()
                    if not h or h in (b"\n", b"\r\n"):
                        break
                    headers.append(h)

                for h in headers:
                    if h.lower().startswith(b"content-length:"):
                        content_length = int(h.split(b":", 1)[1].strip())
                        break

                if content_length and 0 < content_length <= MAX_MESSAGE_SIZE:
                    body = b""
                    while len(body) < content_length:
                        chunk = stdin.read(content_length - len(body))
                        if not chunk:
                            break
                        body += chunk
                    if len(body) == content_length:
                        return body, _MODE_FRAMED
            except Exception:
                continue  # Malformed frame, keep looking

        # 3. If noise, keep looping
        continue


def _send_payload(state, payload: bytes, mode: str) -> None:
    sock = state.get("sock")
    if not sock:
        raise OSError("socket not connected")
    if mode == _MODE_JSONL:
        sock.sendall(payload + b"\n")
    else:
        header = f"Content-Length: {len(payload)}\r\n\r\n".encode("ascii")
        sock.sendall(header + payload)
    trace("proxy_send_payload", bytes=len(payload), mode=mode)


def _reconnect(state) -> bool:
    max_attempts = int(os.environ.get("SARI_PROXY_RECONNECT_MAX", "10") or 10)
    backoff = float(
        os.environ.get(
            "SARI_PROXY_RECONNECT_BACKOFF",
            "0.2") or 0.2)
    with state["conn_lock"]:
        if not state.get("dead") and state.get("sock"):
            return True

        workspace_root = state.get("workspace_root") or ""
        last_err = None
        for _ in range(max_attempts):
            host, port = _resolve_daemon_target()
            if not start_daemon_if_needed(
                    host, port, workspace_root=workspace_root):
                last_err = "daemon start failed"
                time.sleep(backoff)
                continue
            host, port = _resolve_daemon_target()
            try:
                sock = socket.create_connection((host, port))
                last_err = None
                break
            except Exception as e:
                last_err = str(e)
                time.sleep(backoff)
        else:
            _log_error(f"Reconnect failed: {last_err}")
            trace("proxy_reconnect_failed", error=last_err)
            return False

        state["sock"] = sock
        state["dead"] = False
        trace("proxy_reconnect_ok", host=host, port=port)

        t = threading.Thread(
            target=forward_socket_to_stdout, args=(
                sock, state), daemon=True)
        t.start()

        init_req = state.get("init_request")
        if init_req:
            try:
                internal = dict(init_req)
                with state["suppress_lock"]:
                    suppress_ids = state.setdefault("suppress_ids", set())
                    internal_id = -secrets.randbelow(2**31 - 1) - 1
                    while internal_id in suppress_ids:
                        internal_id = -secrets.randbelow(2**31 - 1) - 1
                    suppress_ids.add(internal_id)
                internal["id"] = internal_id
                _send_payload(state, json.dumps(internal).encode(
                    "utf-8"), state.get("mode") or _MODE_FRAMED)
            except Exception as e:
                _log_error(f"Reconnect initialize failed: {e}")
        return True


def forward_stdin_to_socket(state):
    try:
        stdin = sys.stdin.buffer
        while True:
            res = _read_mcp_message(stdin)
            if res is None:
                break
            msg, mode = res
            trace("proxy_received_from_stdin", bytes=len(msg), mode=mode)
            # Inject rootUri for initialize when client omits it
            # (per-connection workspace)
            try:
                req = json.loads(msg.decode("utf-8"))

                def _inject(obj):
                    if not isinstance(obj, dict) or obj.get(
                            "method") != "initialize":
                        return obj, False
                    params = obj.get("params") or {}
                    if params.get("rootUri") or params.get("rootPath"):
                        state["init_request"] = obj
                        return obj, False

                    # PRIORITY: SARI_
                    ws = os.environ.get("SARI_WORKSPACE_ROOT") or state.get(
                        "workspace_root")
                    if not ws:
                        state["init_request"] = obj
                        return obj, False
                    params = dict(params)
                    params["rootUri"] = f"file://{ws}"
                    obj = dict(obj)
                    obj["params"] = params
                    _log_info(f"Injected rootUri for initialize: {ws}")
                    state["init_request"] = obj
                    return obj, True

                injected = False
                if isinstance(req, dict):
                    req, injected = _inject(req)
                elif isinstance(req, list):
                    new_list = []
                    for item in req:
                        item2, did = _inject(item)
                        injected = injected or did
                        new_list.append(item2)
                    req = new_list

                if injected:
                    msg = json.dumps(req).encode("utf-8")
            except Exception:
                pass
            if state.get("mode") is None:
                state["mode"] = mode

            if state.get("dead"):
                if not _reconnect(state):
                    return
            try:
                with state["send_lock"]:
                    _send_payload(state, msg, mode)
            except Exception:
                state["dead"] = True
                if not _reconnect(state):
                    return
                with state["send_lock"]:
                    _send_payload(state, msg, mode)
    except Exception as e:
        _log_error(f"Error forwarding stdin to socket: {e}")
        trace("proxy_forward_stdin_error", error=str(e))
        try:
            sock = state.get("sock")
            if sock:
                sock.close()
        except Exception:
            pass


def _write_error_to_client(message: str, code: int = -32002):
    """Sends a formal JSON-RPC error directly to stdout."""
    resp = {
        "jsonrpc": "2.0",
        "error": {"code": code, "message": message}
    }
    payload = json.dumps(resp).encode("utf-8")
    header = f"Content-Length: {len(payload)}\r\n\r\n".encode("ascii")
    sys.stdout.buffer.write(header + payload)
    sys.stdout.buffer.flush()


def main():
    # Log startup context for diagnostics
    _log_info(
        "Proxy startup: cwd=%s argv=%s SARI_ROOT=%s"
        % (
            os.getcwd(),
            sys.argv,
            os.environ.get("SARI_WORKSPACE_ROOT"),
        )
    )
    trace("proxy_startup", cwd=os.getcwd(), argv=sys.argv,
          workspace_root=os.environ.get("SARI_WORKSPACE_ROOT"))

    workspace_root = os.environ.get(
        "SARI_WORKSPACE_ROOT") or WorkspaceManager.resolve_workspace_root()
    host, port = _resolve_daemon_target()

    if not start_daemon_if_needed(host, port, workspace_root=workspace_root):
        _write_error_to_client(
            "Sari Daemon failed to start. Run 'sari doctor' for diagnostics.")
        sys.exit(1)

    try:
        host, port = _resolve_daemon_target()
        sock = socket.create_connection((host, port))
    except Exception as e:
        err_msg = f"Could not connect to daemon at {host}:{port}: {e}"
        _log_error(err_msg)
        trace("proxy_connect_error", error=str(e))
        _write_error_to_client(err_msg)
        sys.exit(1)

    # Start threads for bidirectional forwarding
    state = {
        "mode": None,
        "sock": sock,
        "dead": False,
        "send_lock": threading.Lock(),
        "conn_lock": threading.Lock(),
        "suppress_lock": threading.Lock(),
        "suppress_ids": set(),
        "init_request": None,
        "workspace_root": workspace_root,
    }
    t1 = threading.Thread(
        target=forward_socket_to_stdout, args=(
            sock, state), daemon=True)
    t1.start()

    forward_stdin_to_socket(state)
    t1.join(timeout=1.0)


if __name__ == "__main__":
    main()



from collections.abc import Mapping
from typing import TypeAlias

from sari.mcp.tools._util import (
    mcp_response,
    pack_header,
    pack_line,
    pack_encode_id,
    pack_encode_text,
    pack_error,
    ErrorCode,
    parse_timestamp,
    invalid_args_response,
)

ToolResult: TypeAlias = dict[str, object]

def execute_archive_context(
    args: object,
    db: object,
    roots: list[str],
    indexer: object = None,
) -> ToolResult:
    """
    도메인 지식이나 작업 컨텍스트를 보관(Archive)하는 도구입니다.
    Facade 패턴을 사용하여 지식 컨텍스트를 DB에 안전하게 저장합니다.
    """
    if not isinstance(args, Mapping):
        return invalid_args_response("archive_context", "args must be an object")

    topic = str(args.get("topic") or "").strip()
    content = str(args.get("content") or "").strip()
    
    if not topic or not content:
        return mcp_response(
            "archive_context",
            lambda: pack_error("archive_context", ErrorCode.INVALID_ARGS, "topic and content are required"),
            lambda: {"error": {"code": ErrorCode.INVALID_ARGS.value, "message": "topic and content are required"}, "isError": True},
        )

    # DTO 기반 업서트를 위한 데이터 준비
    data = {
        "topic": topic,
        "content": content,
        "tags": args.get("tags") or [],
        "related_files": args.get("related_files") or [],
        "source": str(args.get("source") or "").strip(),
        "valid_from": parse_timestamp(args.get("valid_from")),
        "valid_until": parse_timestamp(args.get("valid_until")),
        "deprecated": bool(args.get("deprecated")),
    }

    try:
        # Facade 사용: db.contexts가 모든 내부 세부 사항을 처리합니다.
        payload = db.contexts.upsert(data)
    except Exception as e:
        msg = str(e)
        return mcp_response(
            "archive_context",
            lambda: pack_error("archive_context", ErrorCode.DB_ERROR, msg),
            lambda: {"error": {"code": ErrorCode.DB_ERROR.value, "message": msg}, "isError": True},
        )

    def build_pack() -> str:
        """PACK1 형식의 응답을 생성합니다."""
        lines = [pack_header("archive_context", {"topic": pack_encode_text(payload.topic)}, returned=1)]
        kv = {
            "topic": pack_encode_id(payload.topic),
            "tags": pack_encode_text(",".join(payload.tags)),
        }
        lines.append(pack_line("r", kv))
        return "\n".join(lines)

    return mcp_response("archive_context", build_pack, lambda: payload.model_dump())



# ruff: noqa: E402
import json
import logging
import asyncio
import inspect
import os
import urllib.parse
from typing import Dict,  Optional
from .workspace_registry import Registry, SharedState
from sari.core.settings import settings
from sari.mcp.trace import trace

_SARI_VERSION = settings.VERSION
_SARI_PROTOCOL_VERSION = "2025-11-25"
from sari.core.workspace import WorkspaceManager
from sari.core.server_registry import ServerRegistry

logger = logging.getLogger(__name__)

def _boot_id() -> str:
    return (os.environ.get("SARI_BOOT_ID") or "").strip()

class Session:
    """
    Handles a single client connection.
    Parses JSON-RPC, manages workspace binding via Registry.
    """
    def __init__(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        self.reader = reader
        self.writer = writer
        self.workspace_root: Optional[str] = None
        self.shared_state: Optional[SharedState] = None
        self.registry = Registry.get_instance()
        self.running = True
        self._preinit_server = None
        trace("session_init", has_writer=bool(writer))

    def _get_preinit_server(self):
        if self._preinit_server is None:
            from sari.mcp.server import LocalSearchMCPServer
            workspace_root = WorkspaceManager.resolve_workspace_root()
            self._preinit_server = LocalSearchMCPServer(workspace_root, start_worker=False)
        return self._preinit_server

    async def handle_connection(self):
        trace("session_handle_connection_start")
        try:
            while self.running:
                try:
                    # 1. Read first line to detect framing mode
                    line = await self.reader.readline()
                    if not line: # EOF
                        self.running = False
                        break

                    line_str = line.decode("utf-8").strip()
                    if not line_str:
                        continue

                    request_str = ""
                    if line_str.startswith("{"):
                        # JSONL mode: First line is the JSON body
                        request_str = line_str
                        trace("session_detect_jsonl")
                    elif "content-length:" in line_str.lower():
                        # Content-Length mode: Parse headers
                        headers = {}
                        # Parse first header line
                        parts = line_str.split(":", 1)
                        if len(parts) == 2:
                            headers[parts[0].strip().lower()] = parts[1].strip()
                        
                        # Read remaining headers
                        while True:
                            h_line = await self.reader.readline()
                            if not h_line: # EOF mid-header
                                return 
                            h_str = h_line.decode("utf-8").strip()
                            if not h_str: # End of headers
                                break
                            if ":" in h_str:
                                k, v = h_str.split(":", 1)
                                headers[k.strip().lower()] = v.strip()
                        
                        try:
                            content_length = int(headers.get("content-length", 0))
                        except (ValueError, TypeError):
                            logger.warning(f"Invalid Content-Length: {headers.get('content-length')}")
                            trace("session_invalid_content_length", value=headers.get("content-length"))
                            continue

                        if content_length <= 0:
                            continue

                        body = await self.reader.readexactly(content_length)
                        if not body or len(body) != content_length:
                            logger.warning("Incomplete body read")
                            trace("session_incomplete_body", expected=content_length, actual=len(body) if body else 0)
                            break
                        request_str = body.decode("utf-8")
                        trace("session_read_framed", bytes=content_length)
                    else:
                        # Unknown line, skip but don't crash
                        logger.warning(f"Unknown input line in session: {line_str!r}")
                        trace("session_unknown_line", line_preview=line_str[:120])
                        continue

                    if not request_str:
                        continue

                    try:
                        request = json.loads(request_str)
                        trace("session_request_parsed", keys=sorted(list(request.keys())) if isinstance(request, dict) else type(request).__name__)
                        await self.process_request(request)
                    except json.JSONDecodeError as e:
                        logger.error(f"Invalid JSON received: {e}")
                        trace("session_json_decode_error", error=str(e))
                        await self.send_error(None, -32700, "Parse error")
                    except Exception as e:
                        logger.error(f"Error processing request logic: {e}", exc_info=True)
                        trace("session_process_error", error=str(e))
                        await self.send_error(None, -32603, str(e))

                except (asyncio.IncompleteReadError, ConnectionResetError):
                    logger.info("Connection closed by client (IncompleteRead/Reset)")
                    trace("session_connection_closed")
                    self.running = False
                    break
                except Exception as e:
                    # CRITICAL: Do not let unhandled exceptions kill the loop
                    logger.error(f"Unhandled transport error: {e}", exc_info=True)
                    trace("session_transport_error", error=str(e))
                    # Try to send error if writer is open, but don't crash if it fails
                    try:
                        await self.send_error(None, -32603, "Internal Transport Error")
                    except Exception:
                        pass
                    # Continue loop - do not break

        except Exception as outer_e:
             logger.critical(f"Fatal session error: {outer_e}", exc_info=True)
             trace("session_fatal_error", error=str(outer_e))
        finally:
            self.cleanup()
            try:
                res = self.writer.close()
                if inspect.isawaitable(res):
                    await res
            except Exception:
                pass
            try:
                await self.writer.wait_closed()
            except Exception:
                pass
            trace("session_handle_connection_end")

    async def process_request(self, request: Dict[str, object]):
        method = request.get("method")
        params = request.get("params", {})
        msg_id = request.get("id")
        trace("session_process_request", method=method, msg_id=msg_id, has_shared_state=bool(self.shared_state))

        if self.workspace_root:
            self.registry.touch_workspace(self.workspace_root)

        if method == "sari/identify":
            draining = False
            boot_id = _boot_id()
            latest_info = None
            try:
                reg = ServerRegistry()
                if boot_id:
                    info = reg.get_daemon(boot_id) or {}
                    draining = bool(info.get("draining"))
                
                # Fetch latest non-draining daemon for this workspace
                # (Simple version for Phase 0: just find any latest non-draining daemon)
                daemons = reg.get_active_daemons()
                if daemons:
                    # Sort by version and start time
                    daemons.sort(key=lambda x: (x.get("version", ""), x.get("start_ts", 0)), reverse=True)
                    latest = daemons[0]
                    latest_info = {
                        "host": latest.get("host"),
                        "port": latest.get("port"),
                        "bootId": latest.get("boot_id"),
                        "version": latest.get("version")
                    }
            except Exception:
                draining = False

            response = {
                "jsonrpc": "2.0",
                "id": msg_id,
                "result": {
                    "name": "sari",
                    "version": _SARI_VERSION,
                    "protocolVersion": _SARI_PROTOCOL_VERSION,
                    "bootId": boot_id,
                    "draining": draining,
                    "latest": latest_info
                },
            }
            await self.send_json(response)
            return
        if method == "initialize":
            await self.handle_initialize(request)
        elif method == "initialized":
            # Just forward to server if bound
            if self.shared_state:
                loop = asyncio.get_event_loop()
                await loop.run_in_executor(
                    None,
                    self.shared_state.server.handle_initialized,
                    params
                )
        elif method == "shutdown":
            # Respond to shutdown but keep connection open for exit
            response = {"jsonrpc": "2.0", "id": msg_id, "result": None}
            await self.send_json(response)
        elif method == "exit":
            self.running = False
        else:
            # Forward other requests to the bound server
            if not self.shared_state:
                if method in {"tools/list", "prompts/list", "resources/list", "resources/templates/list", "roots/list", "ping"}:
                    if msg_id is None:
                        return
                    if method == "tools/list":
                        tools = self._get_preinit_server().list_tools()
                        await self.send_json({"jsonrpc": "2.0", "id": msg_id, "result": {"tools": tools}})
                        return
                    if method == "prompts/list":
                        await self.send_json({"jsonrpc": "2.0", "id": msg_id, "result": {"prompts": []}})
                        return
                    if method == "resources/list":
                        await self.send_json({"jsonrpc": "2.0", "id": msg_id, "result": {"resources": []}})
                        return
                    if method == "resources/templates/list":
                        await self.send_json({"jsonrpc": "2.0", "id": msg_id, "result": {"resourceTemplates": []}})
                        return
                    if method == "roots/list":
                        roots = self._get_preinit_server().list_roots()
                        await self.send_json({"jsonrpc": "2.0", "id": msg_id, "result": {"roots": roots}})
                        return
                    if method == "ping":
                        await self.send_json({"jsonrpc": "2.0", "id": msg_id, "result": {}})
                        return
                await self.send_error(msg_id, -32002, "Server not initialized. Send 'initialize' first.")
                trace("session_not_initialized", method=method, msg_id=msg_id)
                return

            # Execute in thread pool to not block async loop
            # Since LocalSearchMCPServer is synchronous
            loop = asyncio.get_event_loop()
            response = await loop.run_in_executor(
                None,
                self.shared_state.server.handle_request,
                request
            )

            if response:
                await self.send_json(response)
                trace("session_response_sent", msg_id=msg_id, method=method)

    async def handle_initialize(self, request: Dict[str, object]):
        params = request.get("params", {})
        msg_id = request.get("id")
        trace("session_handle_initialize_enter", msg_id=msg_id, params_keys=sorted(list(params.keys())))

        boot_id = _boot_id()
        if boot_id:
            try:
                info = ServerRegistry().get_daemon(boot_id) or {}
                if info.get("draining"):
                    await self.send_error(msg_id, -32001, "Server is draining. Reconnect to the latest daemon.")
                    self.running = False
                    return
            except Exception:
                pass

        root_uri = params.get("rootUri") or params.get("rootPath")
        if not root_uri:
            # Fallback for clients that omit rootUri/rootPath
            root_uri = WorkspaceManager.resolve_workspace_root()
        trace("session_handle_initialize_root_uri", root_uri=root_uri)

        if root_uri.startswith("file://"):
            parsed = urllib.parse.urlparse(root_uri)
            if parsed.netloc and parsed.netloc not in {"localhost", "127.0.0.1", "::1"}:
                await self.send_error(msg_id, -32000, "Unsupported file URI host")
                return
            workspace_root = urllib.parse.unquote(parsed.path)
        else:
            workspace_root = root_uri
            
        workspace_root = WorkspaceManager.normalize_path(workspace_root)
        trace("session_handle_initialize_workspace_root", workspace_root=workspace_root)

        # If already bound to a different workspace, release it
        if self.workspace_root and self.workspace_root != workspace_root:
            self.registry.release(self.workspace_root)
            self.shared_state = None

        self.workspace_root = workspace_root
        persist_flag = params.get("sariPersist") or params.get("persist")
        persist = bool(persist_flag) or str(os.environ.get("SARI_PERSIST_WORKSPACE", "")).strip().lower() in {"1", "true", "yes", "on"}
        self.shared_state = self.registry.get_or_create(self.workspace_root, persistent=persist)
        self.registry.touch_workspace(self.workspace_root)
        trace("session_handle_initialize_bound", workspace_root=self.workspace_root)

        # Record workspace mapping in global registry so other clients can find us
        if not boot_id:
            try:
                sockname = self.writer.get_extra_info("sockname") or ("127.0.0.1", 0)
                host = str(sockname[0] or "127.0.0.1")
                port = int(sockname[1] or 0)
                if port > 0:
                    boot_id = f"legacy-{os.getpid()}-{port}"
                    ServerRegistry().register_daemon(boot_id, host, port, os.getpid(), version=_SARI_VERSION)
            except Exception:
                boot_id = ""

        if boot_id:
            try:
                daemon_info = ServerRegistry().get_daemon(boot_id) or {}
                h_port = daemon_info.get("http_port")
                h_host = daemon_info.get("http_host")
                ServerRegistry().set_workspace(workspace_root, boot_id, http_port=h_port, http_host=h_host)
            except Exception:
                try:
                    ServerRegistry().set_workspace(workspace_root, boot_id)
                except Exception:
                    pass

        # Delegate specific initialize logic to the server instance
        # We need to construct the result based on server's response
        # LocalSearchMCPServer.handle_initialize returns the result dict directly
        try:
            result = self.shared_state.server.handle_initialize(params)
            response = {
                "jsonrpc": "2.0",
                "id": msg_id,
                "result": result
            }
            await self.send_json(response)
            trace("session_handle_initialize_success", msg_id=msg_id)
        except Exception as e:
            # Rollback: release the workspace if initialization failed
            self.registry.release(self.workspace_root)
            self.workspace_root = None
            self.shared_state = None
            trace("session_handle_initialize_error", msg_id=msg_id, error=str(e))
            await self.send_error(msg_id, -32000, str(e))

    async def send_json(self, data: Dict[str, object]):
        body = json.dumps(data).encode("utf-8")
        header = f"Content-Length: {len(body)}\r\n\r\n".encode("ascii")
        res = self.writer.write(header + body)
        if inspect.isawaitable(res):
            await res
        await self.writer.drain()
        trace("session_send_json", msg_id=data.get("id"), has_error="error" in data, bytes=len(body))

    async def send_error(self, msg_id: object, code: int, message: str):
        response = {
            "jsonrpc": "2.0",
            "id": msg_id,
            "error": {
                "code": code,
                "message": message
            }
        }
        await self.send_json(response)

    def cleanup(self):
        if self.workspace_root:
            self.registry.release(self.workspace_root)
            # Optional immediate autostop when the last connection closes.
            autostop = settings.get_bool("DAEMON_AUTOSTOP", True)
            if autostop:
                try:
                    if self.registry.active_count() == 0 and not self.registry.has_persistent():
                        boot_id = _boot_id()
                        if boot_id:
                            ServerRegistry().set_daemon_draining(boot_id, True)
                except Exception:
                    pass
            self.workspace_root = None
            self.shared_state = None



"""
MCP client for Sari daemon.

This module handles MCP (Model Context Protocol) communication with the Sari daemon.
"""

import os
import json
import socket
from typing import Optional, TypeAlias

from sari.core.workspace import WorkspaceManager
from sari.core.constants import (
    DAEMON_IDENTIFY_TIMEOUT_SECONDS,
    DAEMON_PROBE_TIMEOUT_SECONDS,
)

JsonMap: TypeAlias = dict[str, object]


def identify_sari_daemon(
    host: str,
    port: int,
    timeout: float = DAEMON_IDENTIFY_TIMEOUT_SECONDS
) -> Optional[JsonMap]:
    """
    Identify if server is a Sari daemon using MCP protocol.

    Tries two methods:
    1. Modern: sari/identify method
    2. Legacy: ping method with "Server not initialized" error

    Args:
        host: Daemon host
        port: Daemon port
        timeout: Connection timeout in seconds

    Returns:
        Identify payload dict if Sari daemon, None otherwise
    """
    # Try modern sari/identify method
    try:
        with socket.create_connection((host, port), timeout=timeout) as sock:
            sock.settimeout(timeout)
            body = json.dumps({"jsonrpc": "2.0", "id": 1,
                              "method": "sari/identify"}).encode("utf-8")
            header = f"Content-Length: {len(body)}\r\n\r\n".encode("ascii")
            sock.sendall(header + body)

            f = sock.makefile("rb")
            headers = {}
            while True:
                line = f.readline()
                if not line:
                    return None
                line = line.strip()
                if not line:
                    break
                if b":" in line:
                    k, v = line.split(b":", 1)
                    headers[k.strip().lower()] = v.strip()

            try:
                content_length = int(headers.get(b"content-length", b"0"))
            except ValueError:
                return None
            if content_length <= 0:
                return None
            resp_body = f.read(content_length)
            if not resp_body:
                return None
            resp = json.loads(resp_body.decode("utf-8"))

            result = resp.get("result") or {}
            if result.get("name") == "sari":
                return result
    except Exception:
        pass

    # Legacy fallback: probe "ping" and accept "Server not initialized" error
    try:
        with socket.create_connection((host, port), timeout=timeout) as sock:
            sock.settimeout(timeout)
            body = json.dumps({"jsonrpc": "2.0", "id": 1,
                              "method": "ping"}).encode("utf-8")
            header = f"Content-Length: {len(body)}\r\n\r\n".encode("ascii")
            sock.sendall(header + body)

            f = sock.makefile("rb")
            headers = {}
            while True:
                line = f.readline()
                if not line:
                    return None
                line = line.strip()
                if not line:
                    break
                if b":" in line:
                    k, v = line.split(b":", 1)
                    headers[k.strip().lower()] = v.strip()

            try:
                content_length = int(headers.get(b"content-length", b"0"))
            except ValueError:
                return None
            if content_length <= 0:
                return None
            resp_body = f.read(content_length)
            if not resp_body:
                return None
            resp = json.loads(resp_body.decode("utf-8"))
            err = resp.get("error") or {}
            msg = (err.get("message") or "").lower()
            if "server not initialized" in msg:
                return {
                    "name": "sari",
                    "version": "legacy",
                    "protocolVersion": ""}
    except Exception:
        pass

    return None


def probe_sari_daemon(
    host: str,
    port: int,
    timeout: float = DAEMON_PROBE_TIMEOUT_SECONDS
) -> bool:
    """
    Verify the server speaks Sari MCP (framed JSON-RPC).

    Args:
        host: Daemon host
        port: Daemon port
        timeout: Connection timeout in seconds

    Returns:
        True if Sari daemon, False otherwise
    """
    return identify_sari_daemon(host, port, timeout=timeout) is not None


def is_http_running(host: str, port: int, timeout: float = 2.0) -> bool:
    """
    Check if Sari HTTP API server is responding.
    """
    import urllib.request
    import json
    try:
        url = f"http://{host}:{port}/health"
        with urllib.request.urlopen(url, timeout=timeout) as r:
            if r.status != 200:
                return False
            payload = json.loads(r.read().decode("utf-8"))
            return bool(payload.get("ok"))
    except Exception:
        return False


def ensure_workspace_http(
    daemon_host: str,
    daemon_port: int,
    workspace_root: Optional[str] = None
) -> bool:
    """
    Ensure workspace is initialized so HTTP server is started/registered.

    Args:
        daemon_host: Daemon host
        daemon_port: Daemon port
        workspace_root: Optional workspace root (auto-detected if None)

    Returns:
        True if successful, False otherwise
    """
    try:
        root = workspace_root or os.environ.get(
            "SARI_WORKSPACE_ROOT") or WorkspaceManager.resolve_workspace_root()
        with socket.create_connection((daemon_host, daemon_port), timeout=1.0) as sock:
            body = json.dumps({
                "jsonrpc": "2.0",
                "id": 1,
                "method": "initialize",
                "params": {
                    "rootUri": f"file://{root}",
                    "capabilities": {},
                    "sariPersist": True,
                },
            }).encode("utf-8")
            header = f"Content-Length: {len(body)}\r\n\r\n".encode("ascii")
            sock.sendall(header + body)
            f = sock.makefile("rb")
            headers = {}
            while True:
                line = f.readline()
                if not line:
                    break
                line = line.strip()
                if not line:
                    break
                if b":" in line:
                    k, v = line.split(b":", 1)
                    headers[k.strip().lower()] = v.strip()
            try:
                content_length = int(headers.get(b"content-length", b"0"))
            except ValueError:
                content_length = 0
            if content_length > 0:
                f.read(content_length)
        return True
    except Exception:
        return False


def request_mcp_status(
    daemon_host: str,
    daemon_port: int,
    workspace_root: Optional[str] = None
) -> Optional[JsonMap]:
    """
    Request status from daemon via MCP.

    Args:
        daemon_host: Daemon host
        daemon_port: Daemon port
        workspace_root: Optional workspace root (auto-detected if None)

    Returns:
        Status dict or None if request fails
    """
    try:
        root = workspace_root or os.environ.get(
            "SARI_WORKSPACE_ROOT") or WorkspaceManager.resolve_workspace_root()
        with socket.create_connection((daemon_host, daemon_port), timeout=2.0) as sock:
            def _send(payload: JsonMap) -> JsonMap:
                body = json.dumps(payload).encode("utf-8")
                header = f"Content-Length: {len(body)}\r\n\r\n".encode("ascii")
                sock.sendall(header + body)
                f = sock.makefile("rb")
                headers = {}
                while True:
                    line = f.readline()
                    if not line:
                        break
                    line = line.strip()
                    if not line:
                        break
                    if b":" in line:
                        k, v = line.split(b":", 1)
                        headers[k.strip().lower()] = v.strip()
                try:
                    content_length = int(headers.get(b"content-length", b"0"))
                except ValueError:
                    content_length = 0
                body = f.read(content_length) if content_length > 0 else b""
                return json.loads(body.decode("utf-8")) if body else {}

            _send({
                "jsonrpc": "2.0",
                "id": 1,
                "method": "initialize",
                "params": {"rootUri": f"file://{root}", "capabilities": {}},
            })
            resp = _send({
                "jsonrpc": "2.0",
                "id": 2,
                "method": "tools/call",
                "params": {"name": "status", "arguments": {"details": True}},
            })
            return resp.get("result") or resp
    except Exception:
        return None



import pytest
import time
import os
import sqlite3
from pathlib import Path
from unittest.mock import MagicMock, patch
from sari.core.server_registry import ServerRegistry
from sari.core.indexer.worker import IndexWorker
from sari.core.db import LocalSearchDB
from sari.mcp.daemon import SariDaemon


def test_registry_workspace_deduplication_policy():
    """
    D1: Verify registry deduplication policy.
    Uses realpath to handle OS-specific temp paths (like /private/tmp on macOS).
    """
    reg = ServerRegistry()
    boot_id = "test-boot-deep"
    # Use realpath to ensure consistency with sari's internal normalization
    parent = os.path.realpath("/tmp/parent_ws")
    child = os.path.realpath("/tmp/parent_ws/src")

    with patch("sari.core.server_registry.get_registry_path") as mock_path:
        reg_file = Path("/tmp/sari_deep_registry.json")
        mock_path.return_value = reg_file
        if reg_file.exists():
            reg_file.unlink()

        reg.register_daemon(boot_id, "127.0.0.1", 47779, os.getpid())

        # 1. Register parent
        reg.set_workspace(parent, boot_id)
        time.sleep(0.01)

        # 2. Register child (more recently active)
        reg.set_workspace(child, boot_id)

        workspaces = reg._load()["workspaces"]

        # Policy Check: Most recent active takes priority in overlap
        assert child in workspaces
        assert parent not in workspaces


def test_registry_prune_dead_tolerates_malformed_daemon_entry():
    """
    D1-1: Malformed daemon payloads should not crash prune logic.
    """
    reg = ServerRegistry()
    data = {
        "version": "2.0",
        "daemons": {
            "bad": 123,  # malformed payload
            "ok": {"pid": os.getpid(), "host": "127.0.0.1", "port": 47779},
        },
        "workspaces": {
            "/tmp/ws-bad": {"boot_id": "bad"},
            "/tmp/ws-ok": {"boot_id": "ok"},
        },
    }

    reg._prune_dead_locked(data)

    assert "bad" not in data["daemons"]
    assert "/tmp/ws-bad" not in data["workspaces"]
    assert "ok" in data["daemons"]
    assert "/tmp/ws-ok" in data["workspaces"]


def test_worker_file_disappeared_mid_process():
    """
    D2: Test worker resilience when file is deleted between stat and read.
    """
    mock_db = MagicMock(spec=LocalSearchDB)
    cfg = MagicMock()
    cfg.store_content = True

    worker = IndexWorker(cfg, mock_db, None, lambda p, c: ([], []))

    root = Path("/tmp/ghost_root")
    file_path = root / "missing.py"
    if file_path.exists():
        file_path.unlink()

    mock_st = MagicMock()
    mock_st.st_mtime = time.time()
    mock_st.st_size = 100

    # worker should catch FileNotFoundError and return None
    res = worker.process_file_task(
        root, file_path, mock_st, int(
            time.time()), time.time(), False)
    assert res is None


def test_db_writer_resilience_to_batch_failure():
    """
    D3: Verify that DBWriter handles batch failures by retrying individually.
    This is a key resilience feature of sari.
    """
    from sari.core.indexer.db_writer import DBWriter, DbTask

    mock_db = MagicMock(spec=LocalSearchDB)
    mock_conn = MagicMock()
    mock_db._write = mock_conn  # DBWriter uses ._write property

    # Mock process_batch to fail on the first call (entire batch)
    # but succeed on subsequent calls (individual tasks)
    writer = DBWriter(mock_db)

    tasks = [
        DbTask(
            kind="upsert_files",
            rows=[
                ("p1",
                 "r1",
                 "id",
                 "repo",
                 0,
                 0,
                 "",
                 "",
                 "",
                 0,
                 0,
                 "ok",
                 "",
                 "ok",
                 "",
                 0,
                 0,
                 0,
                 0,
                 "{}")]),
        DbTask(
            kind="upsert_files",
            rows=[
                ("p2",
                 "r2",
                 "id",
                 "repo",
                 0,
                 0,
                 "",
                 "",
                 "",
                 0,
                 0,
                 "ok",
                 "",
                 "ok",
                 "",
                 0,
                 0,
                 0,
                 0,
                 "{}")])]

    # We mock _process_batch to raise error once, then return success stats
    with patch.object(writer, '_process_batch') as mock_process:
        mock_process.side_effect = [
            sqlite3.OperationalError("database is locked"),  # Batch fails
            {"files": 1},  # Task 1 succeeds
            {"files": 1}  # Task 2 succeeds
        ]

        # Manually run the batch processing logic found in _run loop
        try:
            # Simulate the try-except block in DBWriter._run
            cur = mock_conn.cursor()
            try:
                # 1. Attempt batch
                mock_process(cur, tasks)
                mock_conn.commit()
            except sqlite3.OperationalError:
                # 2. Retry individually (this is the logic we're testing)
                for t in tasks:
                    mock_process(cur, [t])
                    mock_conn.commit()
        except Exception as e:
            pytest.fail(f"Resilience logic failed: {e}")

        # Should be called 3 times: 1 batch + 2 individual retries
        assert mock_process.call_count == 3
        assert mock_conn.commit.call_count == 2  # 2 successful individual commits


@pytest.mark.asyncio
async def test_daemon_duplicate_endpoint_prevention_deep():
    """
    D5: Re-verify daemon prevention with SariDaemon class.
    """
    host = "127.0.0.1"
    port = 49999

    mock_registry = MagicMock(spec=ServerRegistry)
    mock_registry.resolve_daemon_by_endpoint.return_value = {
        "pid": 1, "host": host, "port": port}

    with patch("sari.mcp.daemon.ServerRegistry", return_value=mock_registry):
        daemon = SariDaemon(host=host, port=port)
        with pytest.raises(SystemExit) as excinfo:
            await daemon.start_async()
        assert "already running" in str(excinfo.value)



import sqlite3
import time
import logging

CURRENT_SCHEMA_VERSION = 5
logger = logging.getLogger("sari.db.schema")


def _row_get(row, key: str, index: int, default=None):
    if row is None:
        return default
    try:
        if hasattr(row, "keys"):
            return row[key]
    except Exception:
        pass
    if isinstance(row, (list, tuple)) and len(row) > index:
        return row[index]
    return default


def init_schema(conn: sqlite3.Connection):
    """Sari의 데이터베이스 스키마를 최신 표준에 맞게 초기화하고 마이그레이션을 관리합니다."""
    cur = conn.cursor()

    # 스키마 버전 확인 및 초기 생성
    cur.execute(
        "SELECT name FROM sqlite_master WHERE type='table' AND name='schema_version'")
    if not cur.fetchone():
        _create_all_tables(cur)
        cur.execute("INSERT INTO schema_version (version, applied_ts) VALUES (?, ?)",
                    (CURRENT_SCHEMA_VERSION, int(time.time())))
    else:
        cur.execute("SELECT version FROM schema_version LIMIT 1")
        row = cur.fetchone()
        v = int(_row_get(row, "version", 0, 1) or 1)

        # v2 마이그레이션: 중요도 점수(importance_score) 컬럼 추가
        if v < 2:
            try:
                cur.execute(
                    "ALTER TABLE symbols ADD COLUMN importance_score REAL DEFAULT 0.0")
            except Exception as e:
                logger.debug(
                    "Migration v2 column already exists or failed: %s", e)

        # v3 마이그레이션: 통계 컬럼 및 메타 통계 테이블 추가
        if v < 3:
            try:
                cur.execute(
                    "ALTER TABLE roots ADD COLUMN file_count INTEGER DEFAULT 0")
                cur.execute(
                    "ALTER TABLE roots ADD COLUMN symbol_count INTEGER DEFAULT 0")
                cur.execute(
                    "CREATE TABLE IF NOT EXISTS meta_stats (key TEXT PRIMARY KEY, value TEXT, updated_ts INTEGER)")
            except Exception as e:
                logger.debug("Migration v3 failed: %s", e)

        # v4 마이그레이션: snippet_versions 테이블 추가
        if v < 4:
            try:
                _create_snippet_versions_table(cur)
            except Exception as e:
                logger.debug("Migration v4 failed: %s", e)

        # v5 마이그레이션: symbol_relations 중복 제거 + 유니크 인덱스 추가
        if v < 5:
            try:
                _deduplicate_symbol_relations(cur)
                _create_symbol_relations_indexes(cur)
            except Exception as e:
                logger.debug("Migration v5 failed: %s", e)

        # metadata_json 컬럼 존재 여부 강제 확인 (복구용)
        try:
            cur.execute("SELECT metadata_json FROM files LIMIT 1")
        except Exception:
            try:
                cur.execute(
                    "ALTER TABLE files ADD COLUMN metadata_json TEXT DEFAULT '{}'")
            except Exception as e:
                logger.debug("Failed to add metadata_json column: %s", e)

        # snippet_versions 테이블 존재 여부 강제 확인 (복구용)
        try:
            cur.execute("SELECT id FROM snippet_versions LIMIT 1")
        except Exception:
            try:
                _create_snippet_versions_table(cur)
            except Exception as e:
                logger.debug("Failed to create snippet_versions table: %s", e)

        cur.execute("UPDATE schema_version SET version = ?",
                    (CURRENT_SCHEMA_VERSION,))

    _init_fts(cur)


def _create_all_tables(cur: sqlite3.Cursor):
    """데이터베이스의 모든 테이블을 생성합니다."""
    # 스키마 버전 관리 테이블
    cur.execute(
        "CREATE TABLE IF NOT EXISTS schema_version (version INTEGER PRIMARY KEY, applied_ts INTEGER NOT NULL)")

    _create_roots_table(cur)
    _create_files_table(cur)
    _create_symbols_table(cur)
    _create_symbol_relations_table(cur)
    _create_contexts_table(cur)
    _create_snippets_table(cur)
    _create_snippet_versions_table(cur)
    _create_failed_tasks_table(cur)
    _create_embeddings_table(cur)
    _create_meta_stats_table(cur)


def _create_roots_table(cur: sqlite3.Cursor):
    """워크스페이스 루트 정보를 관리하는 'roots' 테이블을 생성합니다."""
    cur.execute("""
        CREATE TABLE IF NOT EXISTS roots (
            root_id TEXT PRIMARY KEY,    -- 워크스페이스 고유 식별자 (해시 등)
            root_path TEXT NOT NULL,     -- 워크스페이스 원본 경로
            real_path TEXT,              -- 실제 물리적 경로
            last_scan_ts INTEGER DEFAULT 0, -- 마지막 스캔 시간 (Unix Timestamp)
            file_count INTEGER DEFAULT 0,   -- 인덱싱된 파일 수
            symbol_count INTEGER DEFAULT 0, -- 추출된 심볼 수
            config_json TEXT,            -- 워크스페이스별 설정 (JSON)
            label TEXT,                  -- 사용자에게 보여줄 이름
            state TEXT DEFAULT 'ready',  -- 현재 상태 (ready, indexing 등)
            created_ts INTEGER,          -- 생성 일시
            updated_ts INTEGER           -- 수정 일시
        );
    """)


def _create_files_table(cur: sqlite3.Cursor):
    """인덱싱된 파일 정보를 관리하는 'files' 테이블을 생성합니다."""
    cur.execute("""
        CREATE TABLE IF NOT EXISTS files (
            path TEXT PRIMARY KEY,       -- 파일 절대 경로 (식별자)
            rel_path TEXT NOT NULL,      -- 워크스페이스 루트 기준 상대 경로
            root_id TEXT NOT NULL,       -- 소속된 워크스페이스 ID
            repo TEXT,                   -- 소속 저장소 이름 (옵션)
            mtime INTEGER NOT NULL,      -- 파일 수정 시간
            size INTEGER NOT NULL,       -- 파일 크기 (바이트)
            content BLOB,                -- 파일 내용 (선택적 저장/압축 가능)
            hash TEXT,                   -- 파일 내용 해시 (변경 감지용)
            fts_content TEXT,            -- 전체 텍스트 검색(FTS)용 텍스트
            last_seen_ts INTEGER DEFAULT 0, -- 스캔 중 마지막으로 발견된 시간
            deleted_ts INTEGER DEFAULT 0,   -- 삭제된 경우 삭제 일시 (Soft Delete)
            status TEXT DEFAULT 'ok',    -- 분석 상태 (ok, error 등)
            error TEXT,                  -- 분석 중 발생한 오류 내용
            parse_status TEXT DEFAULT 'ok', -- 파싱 상태
            parse_error TEXT,            -- 파싱 오류 내용
            ast_status TEXT DEFAULT 'none', -- AST 추출 상태
            ast_reason TEXT DEFAULT 'none', -- 상태 변화 사유
            is_binary INTEGER DEFAULT 0, -- 이진 파일 여부 (1: true)
            is_minified INTEGER DEFAULT 0, -- 압축(Minified) 파일 여부 (1: true)
            metadata_json TEXT,          -- 추가 메타데이터 (JSON)
            FOREIGN KEY(root_id) REFERENCES roots(root_id)
        );
    """)
    cur.execute("CREATE INDEX IF NOT EXISTS idx_files_root ON files(root_id);")
    cur.execute(
        "CREATE INDEX IF NOT EXISTS idx_files_rel_path ON files(rel_path);")


def _create_symbols_table(cur: sqlite3.Cursor):
    """파일 내 소스코드 심볼(함수, 클래스 등)을 관리하는 'symbols' 테이블을 생성합니다."""
    cur.execute("""
        CREATE TABLE IF NOT EXISTS symbols (
            symbol_id TEXT PRIMARY KEY,  -- 심볼 고유 ID
            path TEXT NOT NULL,          -- 소속 파일 경로
            root_id TEXT NOT NULL,       -- 소속 워크스페이스 ID
            name TEXT NOT NULL,          -- 심볼 이름
            kind TEXT NOT NULL,          -- 심볼 종류 (function, class 등)
            line INTEGER NOT NULL,       -- 정의 시작 라인 (1-based)
            end_line INTEGER NOT NULL,   -- 정의 종료 라인
            content TEXT,                -- 심볼 소스코드 부분
            parent TEXT,                 -- 부모 심볼 이름 (계층 구조)
            meta_json TEXT,              -- 인자 정보 등 추가 메타데이터 (JSON)
            doc_comment TEXT,            -- 관련 문서 주석
            qualname TEXT,               -- 정규화된 이름 (Full Name)
            importance_score REAL DEFAULT 0.0, -- 분석 기반 중요도 점수
            FOREIGN KEY(path) REFERENCES files(path) ON DELETE CASCADE
        );
    """)
    cur.execute("CREATE INDEX IF NOT EXISTS idx_symbols_path ON symbols(path);")
    cur.execute("CREATE INDEX IF NOT EXISTS idx_symbols_name ON symbols(name);")


def _create_symbol_relations_table(cur: sqlite3.Cursor):
    """심볼 간의 상호 참조 관계(호출, 상속 등)를 관리하는 'symbol_relations' 테이블을 생성합니다."""
    cur.execute("""
        CREATE TABLE IF NOT EXISTS symbol_relations (
            from_path TEXT NOT NULL,     -- 호출/참조하는 파일
            from_root_id TEXT NOT NULL,
            from_symbol TEXT NOT NULL,   -- 호출/참조하는 심볼 이름
            from_symbol_id TEXT,         -- 호출/참조하는 심볼 ID (옵션)
            to_path TEXT NOT NULL,       -- 호출/참조 대상 파일
            to_root_id TEXT NOT NULL,
            to_symbol TEXT NOT NULL,     -- 호출/참조 대상 심볼 이름
            to_symbol_id TEXT,           -- 호출/참조 대상 심볼 ID (옵션)
            rel_type TEXT NOT NULL,      -- 관계 종류 (call, inheritance 등)
            line INTEGER,                -- 관계가 발생한 소스 라인
            meta_json TEXT               -- 추가 정보 (JSON)
        );
    """)
    _create_symbol_relations_indexes(cur)


def _create_symbol_relations_indexes(cur: sqlite3.Cursor):
    # Identity-level unique index to prevent repeated accumulation.
    cur.execute(
        """
        CREATE UNIQUE INDEX IF NOT EXISTS ux_symbol_relations_identity
        ON symbol_relations(
            from_path,
            from_root_id,
            from_symbol,
            IFNULL(from_symbol_id, ''),
            to_path,
            to_root_id,
            to_symbol,
            IFNULL(to_symbol_id, ''),
            rel_type,
            IFNULL(line, -1),
            IFNULL(meta_json, '')
        )
        """
    )
    # Query helpers for call graph lookups.
    cur.execute(
        "CREATE INDEX IF NOT EXISTS idx_symbol_relations_to_symbol ON symbol_relations(to_symbol)")
    cur.execute(
        "CREATE INDEX IF NOT EXISTS idx_symbol_relations_to_symbol_id ON symbol_relations(to_symbol_id)")
    cur.execute(
        "CREATE INDEX IF NOT EXISTS idx_symbol_relations_from_symbol ON symbol_relations(from_symbol)")
    cur.execute(
        "CREATE INDEX IF NOT EXISTS idx_symbol_relations_from_symbol_id ON symbol_relations(from_symbol_id)")


def _deduplicate_symbol_relations(cur: sqlite3.Cursor):
    cur.execute(
        """
        DELETE FROM symbol_relations
        WHERE rowid NOT IN (
            SELECT MIN(rowid)
            FROM symbol_relations
            GROUP BY
                from_path,
                from_root_id,
                from_symbol,
                IFNULL(from_symbol_id, ''),
                to_path,
                to_root_id,
                to_symbol,
                IFNULL(to_symbol_id, ''),
                rel_type,
                IFNULL(line, -1),
                IFNULL(meta_json, '')
        )
        """
    )


def _create_contexts_table(cur: sqlite3.Cursor):
    """지적 맥락 정보(Topic 기반 지식 등)를 관리하는 'contexts' 테이블을 생성합니다."""
    cur.execute("""
        CREATE TABLE IF NOT EXISTS contexts (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            topic TEXT UNIQUE,           -- 주제/컨텍스트 키워드
            content TEXT NOT NULL,       -- 컨텍스트 본문
            tags_json TEXT,              -- 관련 태그들 (JSON Array)
            related_files_json TEXT,     -- 연관 파일 정보 (JSON Array)
            source TEXT,                 -- 정보 출처
            valid_from INTEGER,          -- 유효 시작 일시
            valid_until INTEGER,         -- 유효 종료 일시
            deprecated INTEGER DEFAULT 0, -- 폐기 여부 (1: true)
            created_ts INTEGER NOT NULL, -- 생성 일시
            updated_ts INTEGER NOT NULL  -- 수정 일시
        );
    """)


def _create_snippets_table(cur: sqlite3.Cursor):
    """코드 스니펫과 관련 메모를 관리하는 'snippets' 테이블을 생성합니다."""
    cur.execute("""
        CREATE TABLE IF NOT EXISTS snippets (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            tag TEXT NOT NULL,           -- 스니펫 분류/태그
            path TEXT NOT NULL,          -- 소속 파일 경로
            root_id TEXT NOT NULL,       -- 소속 워크스페이스 ID
            start_line INTEGER,          -- 시작 라인
            end_line INTEGER,            -- 종료 라인
            content TEXT,                -- 스니펫 본문 소스
            content_hash TEXT,           -- 본문 해시
            anchor_before TEXT,          -- 주변 맥락 추적용 이전 텍스트
            anchor_after TEXT,           -- 주변 맥락 추적용 이후 텍스트
            repo TEXT,                   -- 소속 저장소
            note TEXT,                   -- 사용자 메모/설명
            commit_hash TEXT,            -- 생성 당시 커밋 해시
            created_ts INTEGER NOT NULL, -- 생성 일시
            updated_ts INTEGER NOT NULL, -- 수정 일시
            metadata_json TEXT,          -- 추가 메타데이터 (JSON)
            FOREIGN KEY(root_id) REFERENCES roots(root_id),
            UNIQUE(tag, root_id, path, start_line, end_line)
        );
    """)


def _create_snippet_versions_table(cur: sqlite3.Cursor):
    """스니펫 이력 관리를 위한 'snippet_versions' 테이블을 생성합니다."""
    cur.execute("""
        CREATE TABLE IF NOT EXISTS snippet_versions (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            snippet_id INTEGER NOT NULL,
            content TEXT NOT NULL,
            content_hash TEXT,
            created_ts INTEGER NOT NULL,
            FOREIGN KEY(snippet_id) REFERENCES snippets(id) ON DELETE CASCADE
        );
    """)
    cur.execute(
        "CREATE INDEX IF NOT EXISTS idx_snippet_versions_snippet_id ON snippet_versions(snippet_id);")


def _create_failed_tasks_table(cur: sqlite3.Cursor):
    """인덱싱 실패 시 재시도 대기를 위한 'failed_tasks' 테이블을 생성합니다."""
    cur.execute("""
        CREATE TABLE IF NOT EXISTS failed_tasks (
            path TEXT PRIMARY KEY,       -- 실패한 파일 경로
            root_id TEXT NOT NULL,       -- 소속 워크스페이스 ID
            attempts INTEGER DEFAULT 0,   -- 지금까지의 재시도 횟수
            error TEXT,                  -- 마지막 에러 메시지
            ts INTEGER,                  -- 발생 일시
            next_retry INTEGER,          -- 다음 재시도 예정 일시
            metadata_json TEXT,          -- 작업 관련 추가 정보 (JSON)
            FOREIGN KEY(root_id) REFERENCES roots(root_id)
        );
    """)


def _create_embeddings_table(cur: sqlite3.Cursor):
    """벡터 검색을 위한 임베딩 데이터를 관리하는 'embeddings' 테이블을 생성합니다."""
    cur.execute("""
        CREATE TABLE IF NOT EXISTS embeddings (
            root_id TEXT NOT NULL,       -- 소속 워크스페이스 ID
            entity_type TEXT NOT NULL,   -- 엔티티 종류 (file, symbol 등)
            entity_id TEXT NOT NULL,     -- 엔티티 ID
            content_hash TEXT,           -- 원본 내용 해시
            model TEXT,                  -- 사용된 모델 이름
            vector BLOB,                 -- 하드웨어 최적화된 벡터 데이터
            created_ts INTEGER,          -- 생성 일시
            updated_ts INTEGER,          -- 수정 일시
            PRIMARY KEY(root_id, entity_type, entity_id)
        );
    """)


def _create_meta_stats_table(cur: sqlite3.Cursor):
    """시스템 메타데이터 및 통계를 보관하는 'meta_stats' 테이블을 생성합니다."""
    cur.execute(
        "CREATE TABLE IF NOT EXISTS meta_stats (key TEXT PRIMARY KEY, value TEXT, updated_ts INTEGER)")


def _init_fts(cur: sqlite3.Cursor):
    """SQLite FTS5를 이용한 고속 텍스트 검색을 초기화합니다."""
    cur.execute(
        "SELECT name FROM sqlite_master WHERE type='table' AND name='files_fts'")
    if cur.fetchone():
        return

    # 외부 콘텐츠(content='files')를 사용하는 가상 테이블 생성
    cur.execute(
        "CREATE VIRTUAL TABLE files_fts USING fts5(path, rel_path, fts_content, content='files', content_rowid='rowid')")

    # files 테이블 변경 시 fts 테이블 자동 동기화를 위한 트리거들
    cur.execute("CREATE TRIGGER files_ai AFTER INSERT ON files BEGIN INSERT INTO files_fts(rowid, path, rel_path, fts_content) VALUES (new.rowid, new.path, new.rel_path, new.fts_content); END")
    cur.execute("CREATE TRIGGER files_ad AFTER DELETE ON files BEGIN INSERT INTO files_fts(files_fts, rowid, path, rel_path, fts_content) VALUES('delete', old.rowid, old.path, old.rel_path, old.fts_content); END")
    cur.execute("CREATE TRIGGER files_au AFTER UPDATE ON files BEGIN INSERT INTO files_fts(files_fts, rowid, path, rel_path, fts_content) VALUES('delete', old.rowid, old.path, old.rel_path, old.fts_content); INSERT INTO files_fts(rowid, path, rel_path, fts_content) VALUES (new.rowid, new.path, new.rel_path, new.fts_content); END")



import os
import importlib
from collections.abc import Mapping
from typing import TypeAlias

from ._util import (
    mcp_response,
    pack_header,
    pack_line,
    pack_encode_id,
    invalid_args_response,
)
try:
    from .call_graph import PLUGIN_API_VERSION
except ImportError:
    PLUGIN_API_VERSION = 1

ToolResult: TypeAlias = dict[str, object]


def _load_plugins() -> list[str]:
    """환경 변수(SARI_CALLGRAPH_PLUGIN)로부터 로드할 호출 그래프 플러그인 목록을 읽어옵니다."""
    mod_path = os.environ.get("SARI_CALLGRAPH_PLUGIN", "").strip()
    if not mod_path:
        return []
    return [m.strip() for m in mod_path.split(",") if m.strip()]

def execute_call_graph_health(
    args: object,
    db: object,
    logger: object = None,
    roots: list[str] | None = None,
) -> ToolResult:
    """
    호출 그래프 플러그인의 상태와 API 호환성을 점검하는 도구입니다.
    로드된 플러그인들의 상태(loaded, error 등)와 버전을 확인합니다.
    """
    if not isinstance(args, Mapping):
        return invalid_args_response("call_graph_health", "args must be an object")

    def build_pack(payload: ToolResult) -> str:
        """PACK1 형식의 응답을 생성합니다."""
        header = pack_header("call_graph_health", {}, returned=1)
        lines = [header]
        for p in payload.get("plugins", []):
            lines.append(pack_line("p", {
                "name": pack_encode_id(p["name"]),
                "status": pack_encode_id(p["status"]),
                "version": str(p.get("version", 0))
            }))
        return "\n".join(lines)

    plugins = _load_plugins()
    results = []
    for p in plugins:
        try:
            # 플러그인 모듈 동적 로드 시도
            mod = importlib.import_module(p)
            results.append({"name": p, "status": "loaded", "version": getattr(mod, "VERSION", PLUGIN_API_VERSION)})
        except Exception as e:
            results.append({"name": p, "status": f"error: {str(e)}", "version": 0})

    payload = {"plugins": results, "api_version": PLUGIN_API_VERSION}
    return mcp_response("call_graph_health", lambda: build_pack(payload), lambda: payload)



import sqlite3
import time
import logging
import os
import zlib
import threading
from typing import Dict, Iterable, List, Optional, Tuple
from peewee import SqliteDatabase
from .models import db_proxy, Root, File
from ..models import ContextDTO, FILE_COLUMNS, SearchOptions, SnippetDTO
from .schema import init_schema
from ..utils.path import PathUtils

logger = logging.getLogger("sari.db")


class LocalSearchDB:
    """
    Sari의 로컬 SQLite 데이터베이스를 관리하는 핵심 클래스입니다.
    파일 인덱스, 심볼, 메타데이터 등을 저장하고 검색하는 기능을 제공합니다.
    Peewee ORM을 기반으로 하며, 성능을 위해 직접 SQL을 실행하거나 메모리 스테이징을 사용하기도 합니다.
    """

    def __init__(self,
                 db_path: str,
                 logger_obj: Optional[logging.Logger] = None,
                 **kwargs):
        """
        Args:
            db_path: SQLite 데이터베이스 파일 경로
            logger_obj: 로거 객체 (기본값: sari.db)
            kwargs: 추가 DB 설정 (journal_mode 등)
        """
        self.db_path = db_path
        self.logger = logger_obj or logger
        self.engine = None
        self._lock = threading.Lock()
        self._writer_thread_id: Optional[int] = None
        try:
            # SQLite 최적화 설정
            self.db = SqliteDatabase(db_path, pragmas={
                # WAL 모드 (동시성 향상)
                'journal_mode': kwargs.get('journal_mode', 'wal'),
                'cache_size': -10000,  # 약 10MB 페이지 캐시
                'foreign_keys': 1,    # 외래 키 제약 조건 활성화
                'ignore_check_constraints': 0,
                'synchronous': 1,     # NORMAL (안전성과 성능의 균형)
                'busy_timeout': 15000  # Lock 대기 시간
            })
            db_proxy.initialize(self.db)
            self.db.connect()
            init_schema(self.db.connection())
            self._init_mem_staging()
        except Exception as e:
            self.logger.error(
                "Failed to initialize database: %s",
                e,
                exc_info=True)
            raise

    def _init_mem_staging(self, force=True):
        """
        대량 삽입 성능을 위해 인메모리 스테이징(staging) 테이블을 초기화합니다.
        메인 디스크 DB에 쓰기 전, 메모리 DB(Attached DB)에 먼저 기록합니다.
        """
        try:
            conn = self.db.connection()
            dbs = [name for _, name, *_ in conn.execute("PRAGMA database_list").fetchall()]
            if "staging_mem" not in dbs:
                conn.execute("ATTACH DATABASE ':memory:' AS staging_mem")
            if force:
                conn.execute("DROP TABLE IF EXISTS staging_mem.files_temp")
            conn.execute(
                "CREATE TABLE IF NOT EXISTS staging_mem.files_temp AS SELECT * FROM main.files WHERE 0")
        except Exception as e:
            self.logger.error(
                "Failed to init mem staging: %s", e, exc_info=True)
            raise

    def _ensure_staging(self):
        try:
            self._init_mem_staging(force=False)
        except Exception:
            self._init_mem_staging(force=True)

    def create_staging_table(self, cur=None):
        self._init_mem_staging(force=True)

    def upsert_root(
            self,
            root_id: str,
            root_path: str,
            real_path: str,
            **kwargs):
        """워크스페이스 루트 정보를 갱신하거나 삽입합니다."""
        with self.db.atomic():
            Root.insert(root_id=root_id,
                        root_path=PathUtils.normalize(root_path),
                        real_path=PathUtils.normalize(real_path),
                        label=kwargs.get("label",
                                         PathUtils.normalize(root_path).split("/")[-1]),
                        updated_ts=int(time.time()),
                        created_ts=int(time.time())).on_conflict_replace().execute()

    def ensure_root(self, root_id: str, path: str):
        self.upsert_root(root_id, path, path)

    def upsert_files_tx(self, cur, rows: List[object]):
        """파일 정보를 트랜잭션 내에서 일괄 삽입합니다 (기본 방식)."""
        self._file_repo(cur).upsert_files_tx(cur, rows)

    def upsert_files_staging(self, cur, rows: List[object]):
        """파일 정보를 메모리 스테이징 테이블에 삽입합니다."""
        self._ensure_staging()
        col_names = ", ".join(FILE_COLUMNS)
        placeholders = ",".join(["?"] * len(FILE_COLUMNS))
        cur.executemany(
            f"INSERT OR REPLACE INTO staging_mem.files_temp({col_names}) VALUES ({placeholders})",
            rows)

    def upsert_files_turbo(self, rows: Iterable[object]):
        """
        대량의 파일 정보를 고속으로 처리합니다 (Turbo Mode).
        데이터를 튜플로 매핑하고 메모리 스테이징 테이블에 기록합니다.
        실제 DB 반영은 finalize_turbo_batch()에서 수행됩니다.
        """
        conn = self.db.connection()
        mapped_tuples = []
        now = int(time.time())
        for r in rows:
            if not r:
                continue
            try:
                if hasattr(r, "to_file_row"):
                    row_tuple = list(r.to_file_row())
                else:
                    row_tuple = list(r)
                while len(row_tuple) < len(FILE_COLUMNS):
                    row_tuple.append(None)
                data = dict(zip(FILE_COLUMNS, row_tuple))
                path = data.get("path")
                if not path:
                    continue
                processed = {
                    "path": PathUtils.normalize(path),
                    "rel_path": data.get("rel_path") or os.path.basename(str(path)),
                    "root_id": data.get("root_id") or "root",
                    "repo": data.get("repo") or "",
                    "mtime": int(data.get("mtime") or now),
                    "size": int(data.get("size") or 0),
                    "content": data.get("content") or b"",
                    "hash": data.get("hash") or "",
                    "fts_content": data.get("fts_content") or "",
                    "last_seen_ts": int(data.get("last_seen_ts") or now),
                    "deleted_ts": int(data.get("deleted_ts") or 0),
                    "status": data.get("status") or "ok",
                    "error": data.get("error"),
                    "parse_status": data.get("parse_status") or "ok",
                    "parse_error": data.get("parse_error"),
                    "ast_status": data.get("ast_status") or "none",
                    "ast_reason": data.get("ast_reason") or "none",
                    "is_binary": int(data.get("is_binary") or 0),
                    "is_minified": int(data.get("is_minified") or 0),
                    "metadata_json": data.get("metadata_json") or "{}"
                }
                mapped_tuples.append(
                    tuple(processed[col] for col in FILE_COLUMNS))
            except Exception as e:
                self.logger.error("Failed to map turbo row: %s", e)
                continue

        if not mapped_tuples:
            return
        try:
            self.upsert_files_staging(conn, mapped_tuples)
            conn.commit()
        except Exception as e:
            self.logger.error(
                "upsert_files_turbo commit failed: %s",
                e,
                exc_info=True)
            raise

    def finalize_turbo_batch(self):
        """메모리 스테이징 테이블의 데이터를 메인 DB 파일로 일괄 이동하고 반영합니다."""
        conn = self.db.connection()
        try:
            res = conn.execute(
                "SELECT count(*) FROM staging_mem.files_temp").fetchone()
            count = int(next(iter(res), 0)) if res else 0
            if count == 0:
                return
            try:
                conn.execute("BEGIN IMMEDIATE TRANSACTION")
                cols = ", ".join(FILE_COLUMNS)
                conn.execute(
                    f"INSERT OR REPLACE INTO main.files({cols}) SELECT {cols} FROM staging_mem.files_temp")
                conn.execute("DELETE FROM staging_mem.files_temp")
                conn.execute("COMMIT")
                self.update_stats()
            except Exception as te:
                self.logger.error(
                    "Database merge failed: %s", te, exc_info=True)
                try:
                    conn.execute("ROLLBACK")
                except Exception:
                    pass
                raise te
        except Exception as e:
            self.logger.error("Critical error in finalize_turbo_batch: %s", e)
            raise

    def update_stats(self):
        """루트별 파일 및 심볼 통계를 갱신합니다."""
        try:
            conn = self.db.connection()
            conn.execute(
                "UPDATE roots SET file_count = (SELECT COUNT(1) FROM files WHERE files.root_id = roots.root_id AND deleted_ts = 0)")
            conn.execute(
                "UPDATE roots SET symbol_count = (SELECT COUNT(1) FROM symbols WHERE symbols.root_id = roots.root_id)")
            conn.commit()
        except Exception as e:
            self.logger.error("Failed to update statistics: %s", e)

    def get_repo_stats(
            self, root_ids: Optional[List[str]] = None) -> Dict[str, int]:
        """레포지토리별 파일 수 통계를 반환합니다."""
        try:
            query = Root.select(Root.label, Root.file_count)
            if root_ids:
                query = query.where(Root.root_id << root_ids)
            return {r.label: r.file_count for r in query}
        except Exception:
            return self._file_repo().get_repo_stats(root_ids=root_ids)

    def execute(self, sql: str, params: Optional[Tuple[object, ...]] = None):
        """얇은 SQL 실행 헬퍼 (상태/진단 경로에서 사용)."""
        conn = self.db.connection()
        if params is None:
            return conn.execute(sql)
        return conn.execute(sql, params)

    def get_roots(self) -> List[Dict[str, object]]:
        """워크스페이스 루트 메타/통계를 대시보드 친화 포맷으로 반환합니다."""
        sql = """
            SELECT
                r.root_id AS root_id,
                r.root_path AS root_path,
                r.real_path AS real_path,
                r.label AS label,
                r.state AS state,
                COALESCE(r.created_ts, 0) AS created_ts,
                COALESCE(r.updated_ts, 0) AS updated_ts,
                COALESCE(r.last_scan_ts, 0) AS last_scan_ts,
                COALESCE(fc.file_count, 0) AS file_count,
                COALESCE(fc.last_indexed_ts, 0) AS last_indexed_ts,
                COALESCE(sc.symbol_count, 0) AS symbol_count
            FROM roots r
            LEFT JOIN (
                SELECT
                    root_id,
                    COUNT(1) AS file_count,
                    MAX(last_seen_ts) AS last_indexed_ts
                FROM files
                WHERE deleted_ts = 0
                GROUP BY root_id
            ) fc ON fc.root_id = r.root_id
            LEFT JOIN (
                SELECT root_id, COUNT(1) AS symbol_count
                FROM symbols
                GROUP BY root_id
            ) sc ON sc.root_id = r.root_id
            ORDER BY r.root_path
        """
        rows = self.execute(sql).fetchall() or []
        results: List[Dict[str, object]] = []
        for row in rows:
            if isinstance(row, sqlite3.Row):
                data = dict(row)
            elif isinstance(row, dict):
                data = dict(row)
            else:
                vals = list(row) if isinstance(row, (list, tuple)) else []
                data = {
                    "root_id": vals[0] if len(vals) > 0 else "",
                    "root_path": vals[1] if len(vals) > 1 else "",
                    "real_path": vals[2] if len(vals) > 2 else "",
                    "label": vals[3] if len(vals) > 3 else "",
                    "state": vals[4] if len(vals) > 4 else "",
                    "created_ts": vals[5] if len(vals) > 5 else 0,
                    "updated_ts": vals[6] if len(vals) > 6 else 0,
                    "last_scan_ts": vals[7] if len(vals) > 7 else 0,
                    "file_count": vals[8] if len(vals) > 8 else 0,
                    "last_indexed_ts": vals[9] if len(vals) > 9 else 0,
                    "symbol_count": vals[10] if len(vals) > 10 else 0,
                }
            root_path = str(data.get("root_path") or "")
            real_path = str(data.get("real_path") or "")
            path = root_path or real_path
            results.append({
                "root_id": str(data.get("root_id") or ""),
                "path": path,
                "root_path": root_path,
                "real_path": real_path,
                "label": str(data.get("label") or ""),
                "state": str(data.get("state") or "ready"),
                "file_count": int(data.get("file_count") or 0),
                "symbol_count": int(data.get("symbol_count") or 0),
                "created_ts": int(data.get("created_ts") or 0),
                "updated_ts": int(data.get("updated_ts") or 0),
                "last_scan_ts": int(data.get("last_scan_ts") or 0),
                "last_indexed_ts": int(data.get("last_indexed_ts") or 0),
            })
        return results

    def read_file(self, path: str) -> Optional[str]:
        """특정 경로의 파일 내용을 DB에서 읽어 반환합니다 (압축 해제 포함)."""
        # Critical: let OperationalError (corruption) bubble up!
        db_path = self._resolve_db_path(path)
        row = File.select(File.content).where(File.path == db_path).first()
        if not row:
            return None
        content = row.content
        if isinstance(content, bytes) and content.startswith(b"ZLIB\0"):
            try:
                content = zlib.decompress(content[5:])
            except Exception as de:
                self.logger.error(
                    "Decompression failed for %s: %s", db_path, de)
                return None
        if isinstance(content, bytes):
            return content.decode("utf-8", errors="ignore")
        return str(content)

    def search_files(self, query: str, limit: int = 50) -> List[Dict]:
        """파일 경로 또는 내용을 기준으로 단순 검색을 수행합니다."""
        lq = f"%{query}%"
        # No try-except here! Let the engine errors be seen.
        return list(File.select().where((File.path ** lq) | (File.rel_path ** lq) |
                    (File.fts_content ** lq)).where(File.deleted_ts == 0).limit(limit).dicts())

    def list_files(self,
                   limit: int = 50,
                   repo: Optional[str] = None,
                   root_ids: Optional[List[str]] = None) -> List[Dict]:
        """조건에 맞는 파일 목록을 반환합니다."""
        return self._file_repo().list_files(limit=limit, repo=repo, root_ids=root_ids)

    def get_file_meta(self, path: str) -> Optional[Tuple[int, int, str]]:
        """파일의 메타데이터(크기, 수정시간, 해시)를 반환합니다."""
        try:
            return self._file_repo().get_file_meta(self._resolve_db_path(path))
        except Exception as e:
            self.logger.debug("Failed to get file meta for %s: %s", path, e)
            return None

    def upsert_symbols_tx(self, cur, rows: List[object], root_id: str = "root"):
        """심볼 정보를 트랜잭션 내에서 일괄 삽입합니다."""
        if not rows:
            return
        if cur is None:
            cur = self.db.connection().cursor()
        self._symbol_repo(cur).upsert_symbols_tx(cur, rows)

    def upsert_relations_tx(self, cur, rows: List[object]):
        """관계 정보를 트랜잭션 내에서 일괄 삽입합니다."""
        if not rows:
            return
        if cur is None:
            cur = self.db.connection().cursor()
        self._symbol_repo(cur).upsert_relations_tx(cur, rows)

    def upsert_snippet_tx(self, cur, rows: List[object]):
        self._snippet_repo(cur).upsert_snippet_tx(cur, rows)

    def list_snippets_by_tag(
            self,
            tag: str,
            limit: int = 20) -> List[SnippetDTO]:
        return self._snippet_repo().list_snippets_by_tag(tag, limit=limit)

    def search_snippets(self, query: str, limit: int = 20) -> List[SnippetDTO]:
        return self._snippet_repo().search_snippets(query, limit=limit)

    def list_snippet_versions(self, snippet_id: int) -> List[Dict[str, object]]:
        return self._snippet_repo().list_snippet_versions(snippet_id)

    def update_snippet_location_tx(
        self,
        cur,
        snippet_id: int,
        start: int,
        end: int,
        content: str,
        content_hash: str,
        anchor_before: str,
        anchor_after: str,
        updated_ts: int,
    ) -> None:
        self._snippet_repo(cur).update_snippet_location_tx(
            cur,
            snippet_id,
            start,
            end,
            content,
            content_hash,
            anchor_before,
            anchor_after,
            updated_ts,
        )

    def upsert_context_tx(self, cur, rows: List[object]):
        self._context_repo(cur).upsert_context_tx(cur, rows)

    def search_contexts(self, query: str, limit: int = 20) -> List[ContextDTO]:
        return self._context_repo().search_contexts(query, limit=limit)

    def prune_stale_data(self, root_id: str, active_paths: List[str]):
        """더 이상 존재하지 않는 파일 데이터를 DB에서 정리(제거)합니다."""
        with self.db.atomic():
            if active_paths:
                File.delete().where(
                    (File.root == root_id) & (
                        File.path.not_in(active_paths))).execute()
            else:
                File.delete().where(File.root == root_id).execute()

    def delete_path_tx(self, cur, path: str):
        self._file_repo(cur).delete_path_tx(cur, path)

    def update_last_seen_tx(self, cur, paths: List[str], ts: int):
        self._file_repo(cur).update_last_seen_tx(cur, paths, ts)

    def search_symbols(
            self,
            query: str,
            limit: int = 20,
            **kwargs) -> List[Dict]:
        """심볼 이름으로 검색을 수행합니다."""
        return [
            s.model_dump() for s in self._symbol_repo().search_symbols(
                query, limit=limit, **kwargs)]

    def search_v2(self, opts: SearchOptions):
        """V2 검색 인터페이스 (엔진 위임 또는 저장소 직접 검색)."""
        if self.engine and hasattr(self.engine, "search_v2"):
            return self.engine.search_v2(opts)
        return self._search_repo().search_v2(opts)

    def repo_candidates(self, q: str, limit: int = 3,
                        root_ids: Optional[List[str]] = None) -> List[Dict[str, object]]:
        if self.engine and hasattr(self.engine, "repo_candidates"):
            return self.engine.repo_candidates(
                q, limit=limit, root_ids=root_ids)
        return self._search_repo().repo_candidates(q, limit=limit, root_ids=root_ids)

    def search_sqlite_v2(self, opts: SearchOptions):
        """SQLite-only fallback search path (engine bypass)."""
        return self._search_repo().search_v2(opts)

    def repo_candidates_sqlite(self, q: str, limit: int = 3,
                               root_ids: Optional[List[str]] = None) -> List[Dict[str, object]]:
        """SQLite-only repo candidate path (engine bypass)."""
        return self._search_repo().repo_candidates(q, limit=limit, root_ids=root_ids)

    def apply_root_filter(
            self, sql: str, root_id: Optional[str]) -> Tuple[str, List[object]]:
        sql = str(sql or "").strip()
        if not sql:
            return sql, []
        has_where = " where " in sql.lower()
        params: List[object] = []
        if root_id:
            if has_where:
                sql += " AND root_id = ?"
            else:
                sql += " WHERE root_id = ?"
            params.append(str(root_id))
        elif not has_where:
            sql += " WHERE 1=1"
        return sql, params

    def count_failed_tasks(self) -> Tuple[int, int]:
        return self.tasks.count_failed_tasks()

    def register_writer_thread(self, thread_id: Optional[int]) -> None:
        self._writer_thread_id = int(
            thread_id) if thread_id is not None else None

    def get_symbol_fan_in_stats(
            self, symbol_names: List[str]) -> Dict[str, int]:
        """심볼의 참조 횟수(Fan-in) 통계를 반환합니다."""
        return self._symbol_repo().get_symbol_fan_in_stats(symbol_names)

    def has_legacy_paths(self) -> bool: return False
    def set_engine(self, engine): self.engine = engine

    def swap_db_file(self, new_path: str):
        """
        워커 프로세스에서 생성된 스냅샷 DB의 내용을 메인 DB로 병합합니다.
        ATTACH DATABASE를 사용하여 테이블 간 데이터 복사를 수행합니다.
        """
        if not new_path or not os.path.exists(new_path):
            return
        conn = self.db.connection()
        try:
            conn.execute("ATTACH DATABASE ? AS snapshot", (new_path,))
            with self.db.atomic():
                for tbl in [
                    "roots",
                    "files",
                    "symbols",
                    "symbol_relations",
                    "snippets",
                    "failed_tasks",
                        "embeddings"]:
                    try:
                        if tbl == "files":
                            cols = ", ".join(FILE_COLUMNS)
                            conn.execute(
                                f"INSERT OR REPLACE INTO main.files({cols}) SELECT {cols} FROM snapshot.files")
                        else:
                            conn.execute(
                                f"INSERT OR REPLACE INTO main.{tbl} SELECT * FROM snapshot.{tbl}")
                    except Exception as te:
                        self.logger.error(
                            "Failed to copy table %s from snapshot: %s", tbl, te)
            self.update_stats()
        except Exception as e:
            self.logger.error("Failed to swap DB file: %s", e, exc_info=True)
            raise
        finally:
            try:
                conn.execute("DETACH DATABASE snapshot")
            except Exception as de:
                self.logger.debug("Failed to detach snapshot: %s", de)

    def get_connection(self): return self.db.connection()

    def get_read_connection(self):
        conn = self.db.connection()
        conn.row_factory = sqlite3.Row
        return conn

    @property
    def _write(self): return self.db.connection()
    @property
    def _read(self): return self.get_read_connection()
    def _get_conn(self): return self.db.connection()

    # 하위 리포지토리 접근 프로퍼티 (지연 로딩)
    @property
    def files(self): return self._file_repo()
    @property
    def symbols(self): return self._symbol_repo()
    @property
    def snippets(self): return self._snippet_repo()
    @property
    def search_repo(self): return self._search_repo()

    @property
    def tasks(self):
        from sari.core.repository.failed_task_repository import FailedTaskRepository
        return FailedTaskRepository(self.db.connection())

    @property
    def contexts(self):
        return self._context_repo()

    def _resolve_db_path(self, path: str) -> str:
        """절대 경로를 DB 내부의 상대 경로(root_id/rel_path)로 변환합니다."""
        if os.path.isabs(path):
            from sari.core.workspace import WorkspaceManager
            root = WorkspaceManager.find_root_for_path(path)
            if root:
                rid = WorkspaceManager.root_id(root)
                rel = PathUtils.to_relative(path, root)
                return f"{rid}/{rel}"
        return path

    def _file_repo(self, cur=None):
        from sari.core.repository.file_repository import FileRepository
        return FileRepository(self._get_real_conn(cur))

    def _symbol_repo(self, cur=None):
        from sari.core.repository.symbol_repository import SymbolRepository
        return SymbolRepository(self._get_real_conn(cur))

    def _search_repo(self, cur=None):
        from sari.core.repository.search_repository import SearchRepository
        return SearchRepository(self._get_real_conn(cur))

    def _snippet_repo(self, cur=None):
        from sari.core.repository.extra_repository import SnippetRepository
        return SnippetRepository(self._get_real_conn(cur))

    def _context_repo(self, cur=None):
        from sari.core.repository.extra_repository import ContextRepository
        return ContextRepository(self._get_real_conn(cur))

    def _get_real_conn(self, cur):
        if cur is not None:
            if hasattr(cur, "connection"):
                return cur.connection
            if hasattr(cur, "execute"):
                return cur
        return self.db.connection()

    def close(self): self.db.close()
    def close_all(self): self.close()



from types import SimpleNamespace

from sari.core.models import SearchHit
from sari.core.services.search_service import SearchService


def test_search_service_hybrid_tolerates_non_mapping_meta():
    keyword_hits = [SearchHit(path="a.py", hit_reason="keyword")]
    semantic_hits = [SearchHit(path="b.py", hit_reason="semantic")]

    class _Engine:
        @staticmethod
        def search_v2(_opts):
            return keyword_hits, None

    class _Repo:
        @staticmethod
        def search_semantic(_query_vector, limit=10, root_ids=None):
            return semantic_hits

    class _DB:
        @staticmethod
        def search_repo():
            return _Repo()

    svc = SearchService(_DB(), _Engine(), indexer=None)
    opts = SimpleNamespace(limit=10, root_ids=[], query_vector=[0.1, 0.2])

    hits, meta = svc.search(opts)

    assert len(hits) == 2
    assert meta["engine"] == "hybrid-rrf"


def test_search_service_index_meta_fallback_when_to_meta_is_not_mapping():
    class _Status:
        @staticmethod
        def to_meta():
            return ["bad-meta"]

        index_ready = True
        indexed_files = 3
        scanned_files = 5
        errors = 0
        symbols_extracted = 7
        index_version = "v1"
        last_error = ""
        scan_started_ts = 10
        scan_finished_ts = 20

    svc = SearchService(db=None, engine=None, indexer=SimpleNamespace(status=_Status()))
    meta = svc.index_meta()

    assert isinstance(meta, dict)
    assert meta["index_ready"] is True
    assert meta["indexed_files"] == 3



from sari.core.db.main import LocalSearchDB
from sari.core.db.storage import GlobalStorageManager

def test_storage_direct_turbo_persistence(tmp_path):
    """
    Verify that the storage backend correctly persists data through the Turbo path.
    """
    db_path = tmp_path / "storage.db"
    db = LocalSearchDB(str(db_path))
    db.ensure_root("root", str(tmp_path))

    # 1. Prepare bulk data
    rows = []
    for i in range(100):
        rows.append((f"p{i}", f"rel{i}", "root", "repo", 0, 10, b"data", "h", "fts", 0, 0, "ok", "", "ok", "", 0, 0, 0, 10, "{}"))

    # 2. Bulk Ingestion
    db.upsert_files_turbo(rows)
    db.finalize_turbo_batch()

    # 3. Persistence Check
    assert len(db.search_files("rel50")) == 1
    db.close()

    # 4. Reload Check (Real durability)
    db2 = LocalSearchDB(str(db_path))
    assert len(db2.search_files("rel99")) == 1


def test_storage_overlay_mtime_respects_newer_rows():
    class DummyDB:
        db_path = ":memory:"

    storage = GlobalStorageManager(DummyDB())

    path = "root/rel/path.py"
    row_new = (path, "rel/path.py", "root", "repo", 100, 10, b"", "h", "snippet-new")
    row_old = (path, "rel/path.py", "root", "repo", 90, 10, b"", "h", "snippet-old")
    row_newer = (path, "rel/path.py", "root", "repo", 110, 10, b"", "h", "snippet-newer")

    storage.upsert_files([row_new])
    storage.upsert_files([row_old])
    assert storage._overlay_files[path]["mtime"] == 100
    assert storage._overlay_files[path]["snippet"] == "snippet-new"

    storage.upsert_files([row_newer])
    assert storage._overlay_files[path]["mtime"] == 110
    assert storage._overlay_files[path]["snippet"] == "snippet-newer"
